from rest_framework import serializers
from arkindex.project.serializer_fields import EnumField
from arkindex.project.aws import S3FileStatus
from arkindex.dataimport.models import DataFile
from arkindex.documents.models import Corpus
from arkindex.images.serializers import ImageSerializer


class DataFileSerializer(serializers.ModelSerializer):
    """
    Serialize a single uploaded file
    """

    images = ImageSerializer(many=True, read_only=True)
    status = EnumField(S3FileStatus)
    s3_url = serializers.SerializerMethodField()

    class Meta:
        model = DataFile
        fields = (
            'id',
            'name',
            'hash',
            'content_type',
            'size',
            'images',
            'status',
            's3_url',
        )
        read_only_fields = ('id', 'name', 'hash', 'size', 'content_type', 'images', 's3_url', )

    def validate_status(self, value):
        if value == S3FileStatus.Checked:
            # Status has been resquested to be checked, perform validation
            try:
                self.instance.perform_check(raise_exc=True)
            except (AssertionError, ValueError) as e:
                raise serializers.ValidationError(str(e))
        return value

    def get_s3_url(self, obj):
        if 'request' not in self.context:
            return
        # Only allow the S3 URL for internal users or admins
        user = self.context['request'].user
        if user.is_authenticated and (user.is_admin or user.is_internal):
            return obj.s3_url


class DataFileCreateSerializer(serializers.ModelSerializer):
    """
    Serialize a Datafile creation with Amazon S3 PUT uri
    """

    status = EnumField(S3FileStatus, read_only=True)
    s3_put_url = serializers.SerializerMethodField()
    corpus = serializers.PrimaryKeyRelatedField(queryset=Corpus.objects.none())

    class Meta:
        model = DataFile
        fields = (
            'id',
            'name',
            'hash',
            'size',
            'corpus',
            'status',
            's3_url',
            's3_put_url',
        )
        read_only_fields = ('id', 'status', 's3_url', 's3_put_url')
        # Remove the default UniqueTogetherValidator
        validators = []

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.context.get('request'):
            # Do not raise Error in order to create OpenAPI schema
            return
        self.fields['corpus'].queryset = Corpus.objects.writable(self.context['request'].user)

    def get_s3_put_url(self, obj):
        if obj.status == S3FileStatus.Checked:
            return None
        return obj.s3_put_url

    def validate(self, data):
        existing_datafile = DataFile.objects.filter(hash=data['hash'], corpus_id=data['corpus']).first()
        if existing_datafile:
            message = {
                'hash': ['DataFile with this hash already exists in corpus'],
                'id': str(existing_datafile.id),
                'status': existing_datafile.status.value,
            }
            if existing_datafile.status != S3FileStatus.Checked:
                message['s3_put_url'] = existing_datafile.s3_put_url
            else:
                message['s3_url'] = existing_datafile.s3_url
            self._errors = message
            raise serializers.ValidationError(message)

        return super().validate(data)
