from rest_framework import serializers
from rest_framework.utils import model_meta
from ponos.models import State
from arkindex_common.enums import DataImportMode, DataImportPDFEngine
from arkindex.project.serializer_fields import EnumField
from arkindex.documents.models import Corpus, Element, ElementType
from arkindex.dataimport.models import DataImport, DataFile
from arkindex.dataimport.serializers.git import RevisionSerializer
from arkindex.dataimport.serializers.ml_tool import MLToolTaskSerializer


class DataImportLightSerializer(serializers.ModelSerializer):
    """
    Serialize a data importing workflow
    """

    state = EnumField(State)
    mode = EnumField(DataImportMode)
    creator = serializers.HiddenField(default=serializers.CurrentUserDefault())

    class Meta:
        model = DataImport
        fields = (
            'id',
            'state',
            'mode',
            'corpus',
            'creator',
        )
        read_only_fields = ('id', 'state')


class ImagesPayloadSerializer(serializers.Serializer):
    """
    Serialize an image importing payload
    """

    folder_name = serializers.CharField()
    volume_id = serializers.CharField()
    ml_tools = MLToolTaskSerializer(many=True, required=False, default=list)


class PDFPayloadSerializer(ImagesPayloadSerializer):
    """
    Serialize a pdf importing payload
    """

    pdf_engine = EnumField(DataImportPDFEngine, default=DataImportPDFEngine.Convert)


class DataImportSerializer(DataImportLightSerializer):
    """
    Serialize a data importing workflow with its payload
    """

    # Redefine state as writable
    state = EnumField(State)
    payload = serializers.JSONField()
    revision = RevisionSerializer(read_only=True)
    workflow = serializers.HyperlinkedRelatedField(read_only=True, view_name='ponos:workflow-details')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.instance is None:
            return
        payload_serializers = {
            DataImportMode.Images: ImagesPayloadSerializer,
            DataImportMode.PDF: PDFPayloadSerializer,
        }
        self.fields['payload'] = payload_serializers.get(self.instance.mode, serializers.JSONField)()

    def update(self, instance, validated_data):
        # ModelSerializer disallows updating with nested serializers.
        # We dynamically inject a payload serializer, but it is useless for the database
        # since the payload field is a JSONField with no particular checks.
        # We don't have to care about a nested serializer and we can treat it just as a dict.
        # This code is copy-pasted from DRF's ModelSerializer.
        info = model_meta.get_field_info(instance)
        for attr, value in validated_data.items():
            if attr in info.relations and info.relations[attr].to_many:
                field = getattr(instance, attr)
                field.set(value)
            else:
                setattr(instance, attr, value)
        instance.save()
        return instance

    class Meta(DataImportLightSerializer.Meta):
        fields = DataImportLightSerializer.Meta.fields + (
            'payload',
            'files',
            'revision',
            'workflow',
        )


class DataImportFromFilesSerializer(serializers.Serializer):

    mode = EnumField(DataImportMode, default=DataImportMode.Images)
    files = serializers.PrimaryKeyRelatedField(queryset=DataFile.objects.all(), many=True)
    volume_id = serializers.UUIDField(required=False, allow_null=True)
    volume_name = serializers.CharField(max_length=250, required=False, allow_null=True)
    pdf_engine = EnumField(DataImportPDFEngine, default=DataImportPDFEngine.Convert)
    ml_tools = MLToolTaskSerializer(many=True, required=False, allow_null=True)

    def validate_mode(self, mode):
        if mode not in (DataImportMode.Images, DataImportMode.PDF):
            raise serializers.ValidationError("This mode is not allowed when importing from files")
        return mode

    def validate_files(self, files):
        if not files:
            raise serializers.ValidationError('At least one file is required to start an import')
        corpora = set(f.corpus for f in files)
        if len(corpora) > 1:
            raise serializers.ValidationError('Imports can only run on files from a single corpus')
        corpus = corpora.pop()
        if corpus not in Corpus.objects.writable(self.context['request'].user):
            raise serializers.ValidationError('Cannot write in corpus')
        return files

    def validate_volume_id(self, volume_id):
        if not volume_id:
            return None
        try:
            volume = Element.objects.get(
                id=volume_id,
                type=ElementType.Volume,
            )
            return volume
        except Element.DoesNotExist:
            raise serializers.ValidationError('Volume does not exist')

    def validate(self, data):
        if data['mode'] == DataImportMode.PDF:
            if len(data['files']) > 1:
                raise serializers.ValidationError('Cannot import more than one PDF at once')
            if data['files'][0].content_type != 'application/pdf':
                raise serializers.ValidationError('PDF imports can only import PDF files')

        elif data['mode'] == DataImportMode.Images:
            if not all(f.content_type.startswith('image/') for f in data['files']):
                raise serializers.ValidationError('Image imports can only import image files')

        else:
            raise NotImplementedError

        return data
