from unittest.mock import MagicMock, patch, call
from arkindex_common.enums import DataImportMode
from arkindex.project.tests import FixtureTestCase
from arkindex.documents.models import Element, Page
from arkindex.dataimport.models import DataImport
from arkindex.dataimport.filetypes import IIIFFileType, SurfaceFileType
from arkindex.dataimport.git import GitFlow, SimpleDiff, DiffType
from ponos.models import State
import os.path
import tempfile
import shutil

FIXTURES = os.path.join(
    os.path.dirname(os.path.realpath(__file__)),
    'manifest_samples',
)


class TestGitFlow(FixtureTestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.working_dir = tempfile.mkdtemp()

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()
        shutil.rmtree(cls.working_dir)

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.repo = cls.corpus.repos.get()
        cls.rev = cls.repo.revisions.get()

    def setUp(self):
        self.dataimport = DataImport(
            corpus=self.corpus,
            creator=self.user,
            revision=self.rev,
            payload={
                "repo_id": str(self.repo.id),
                "sha": self.rev.hash,
            },
            mode=DataImportMode.Repository,
        )

    def test_mode(self):
        """
        Test GitFlow checks the DataImportMode
        """
        self.dataimport.mode = DataImportMode.Images
        with self.assertRaises(AssertionError):
            GitFlow(self.dataimport, self.working_dir)

    def test_repo(self):
        """
        Test GitFlow checks the DataImport has a repository
        """
        self.dataimport.revision = None
        del self.dataimport.payload['repo_id']
        self.assertIsNone(self.dataimport.repo)
        with self.assertRaises(AssertionError):
            GitFlow(self.dataimport, self.working_dir)

    def test_sha(self):
        """
        Test GitFlow checks for a revision SHA in DataImport
        """
        self.dataimport.revision = None
        del self.dataimport.payload['sha']
        self.assertIsNone(self.dataimport.sha)
        with self.assertRaises(AssertionError):
            GitFlow(self.dataimport, self.working_dir)

    @patch('arkindex.dataimport.models.get_provider')
    def test_provider(self, get_provider_mock):
        """
        Test GitFlow checks for a repository provider in DataImport
        """
        get_provider_mock.return_value = None
        self.assertIsNone(self.dataimport.repo.provider_class)
        with self.assertRaisesRegex(ValueError, 'provider'):
            GitFlow(self.dataimport, self.working_dir)

    def test_directory_exists(self):
        """
        Test GitFlow checks the working directory exists
        """
        with self.assertRaisesRegex(AssertionError, 'working directory'):
            GitFlow(self.dataimport, '/does/not/exist')

    @patch('arkindex.dataimport.git.ConfigFile')
    @patch('arkindex.dataimport.models.get_provider')
    def test_parse(self, provider_mock, config_mock):
        """
        Test GitFlow performs config parsing
        """
        self.dataimport.revision = None
        provider_mock()().get_file_content.return_value = 'test'
        provider_mock()().get_or_create_revision.return_value = (self.rev, True)
        self.assertFalse(config_mock.called)
        self.assertIsNone(self.dataimport.revision)

        config = GitFlow(self.dataimport, self.working_dir).parse()

        self.assertEqual(config, config_mock.return_value)
        self.assertEqual(self.dataimport.revision, self.rev)
        self.assertTrue(config_mock.call_count, 1)
        self.assertTrue(config_mock.call_args, call('test'))

    @patch('arkindex.dataimport.models.get_provider')
    def test_parse_remote_errors(self, provider_mock):
        """
        Test GitFlow handles remote errors when fetching config data
        """
        provider_mock()().get_file_content.side_effect = Exception
        with self.assertRaisesRegex(Exception, 'fetching configuration'):
            GitFlow(self.dataimport, self.working_dir).parse()

    @patch('arkindex.dataimport.git.ConfigFile')
    @patch('arkindex.dataimport.models.get_provider')
    def test_parse_errors(self, provider_mock, config_mock):
        """
        Test GitFlow handles config parsing errors
        """
        provider_mock()().get_file_content.return_value = ''
        config_mock.side_effect = AssertionError
        with self.assertRaisesRegex(ValueError, 'validation failed'):
            GitFlow(self.dataimport, self.working_dir).parse()

    @patch('arkindex.dataimport.models.get_provider')
    def test_clone(self, provider_mock):
        """
        Test a normal GitFlow cloning
        """
        repo = GitFlow(self.dataimport, self.working_dir).clone()
        self.assertEqual(provider_mock()().clone_repo.return_value, repo)
        self.assertTrue(provider_mock()().clone_repo.call_count, 1)
        self.assertEqual(provider_mock()().clone_repo.call_args, call(
            self.repo,
            os.path.join(self.working_dir, 'repo'),
            no_checkout=True,
        ))
        self.assertEqual(provider_mock()().clone_repo().create_head.call_count, 1)
        self.assertEqual(
            provider_mock()().clone_repo().create_head.call_args,
            call('commit_42', '42'),
        )
        self.assertEqual(provider_mock()().clone_repo().head.reset.call_count, 1)
        self.assertEqual(
            provider_mock()().clone_repo().head.reset.call_args,
            call(index=True, working_tree=True),
        )

    @patch('arkindex.dataimport.git.os.path.exists')
    def test_clone_exists(self, exists_mock):
        """
        Test GitFlow checks the repo does not already exist
        """
        exists_mock.return_value = True
        with self.assertRaisesRegex(AssertionError, 'Already existing'):
            GitFlow(self.dataimport, self.working_dir).clone()
        self.assertEqual(exists_mock.call_count, 1)
        self.assertEqual(exists_mock.call_args, call(os.path.join(self.working_dir, 'repo')))

    @patch('arkindex.dataimport.models.get_provider')
    def test_clone_errors(self, provider_mock):
        """
        Test GitFlow handles cloning errors
        """
        provider_mock()().clone_repo.side_effect = Exception
        with self.assertRaisesRegex(Exception, 'error occured while cloning'):
            GitFlow(self.dataimport, self.working_dir).clone()

    @patch('arkindex.dataimport.models.get_provider')
    def test_clone_checkout_errors(self, provider_mock):
        """
        Test GitFlow handles checkout errors
        """
        provider_mock()().clone_repo().head.reset.side_effect = Exception
        with self.assertRaisesRegex(Exception, 'error occured while checking out'):
            GitFlow(self.dataimport, self.working_dir).clone()

    def test_diff_from_scratch(self):
        """
        Test GitFlow can return add-only diffs when there are no known revisions
        """
        flow = GitFlow(self.dataimport, self.working_dir)
        flow.repo = MagicMock()
        flow.repo.iter_commits.return_value = []
        flow.repo.git.ls_files.return_value = 'path1\npath2'
        diffs = flow.diff()
        self.assertListEqual(diffs, [
            SimpleDiff(DiffType.Addition, 'path1', 'path1'),
            SimpleDiff(DiffType.Addition, 'path2', 'path2'),
        ])
        self.assertEqual(flow.repo.iter_commits.call_count, 1)
        self.assertEqual(flow.repo.git.ls_files.call_count, 1)

    def test_diff_revision(self):
        """
        Test GitFlow performs a diff when there is an existing revision
        """
        new_rev = self.repo.revisions.create(hash='1337', message='a', ref='master', author='me')
        self.dataimport.start()
        self.dataimport.workflow.tasks.all().update(state=State.Completed)
        self.assertEqual(self.dataimport.state, State.Completed)
        self.dataimport.id = None
        self.dataimport.revision = new_rev

        diff1, diff2 = MagicMock(), MagicMock()
        diff1.change_type, diff2.change_type = 'M', 'D'
        diff1.a_path, diff1.b_path = 'path1a', 'path1b'
        diff2.a_path, diff2.b_path = 'path2a', 'path2b'

        commit1, commit2 = MagicMock(), MagicMock()
        commit1.hexsha = '42'
        commit2.hexsha = '1337'
        commit1.diff.return_value = [diff1, diff2]

        repo = MagicMock()
        repo.commit.return_value = commit2
        repo.iter_commits.return_value = [commit1, commit2]

        flow = GitFlow(self.dataimport, self.working_dir)
        flow.repo = repo
        diffs = flow.diff()

        self.assertListEqual(diffs, [
            SimpleDiff(DiffType.Modification, 'path1a', 'path1b'),
            SimpleDiff(DiffType.Deletion, 'path2a', 'path2b'),
        ])

        self.assertEqual(commit1.diff.call_count, 1)
        self.assertEqual(commit1.diff.call_args, call(commit2))
        self.assertEqual(repo.commit.call_count, 1)
        self.assertEqual(repo.commit.call_args, call('1337'))
        self.assertEqual(repo.iter_commits.call_count, 1)
        self.assertEqual(repo.iter_commits.call_args, call('1337'))

    def test_diff_check_workflows(self):
        """
        Test GitFlow performs a diff with existing revisions with successful imports

        Example with commits "42 -> 1337 -> cafe":
        Importing rev 42: None -> 42, completed
        Importing rev 1337: 42 -> 1337, failed
        Expected diff when importing rev cafe: 42 -> cafe
        """
        # Setup Revisions
        rev_1337 = self.repo.revisions.create(hash='1337', message='a', ref='master', author='me')
        rev_cafe = self.repo.revisions.create(hash='cafe', message='tasty', ref='master', author='me')

        # Setup workflows
        self.dataimport.start()
        self.dataimport.workflow.tasks.all().update(state=State.Completed)
        self.assertEqual(self.dataimport.state, State.Completed)
        import1337 = DataImport(
            corpus=self.corpus,
            creator=self.user,
            revision=rev_1337,
            payload={
                "repo_id": str(self.repo.id),
                "sha": rev_1337.hash,
            },
            mode=DataImportMode.Repository,
        )
        import1337.start()
        import1337.workflow.tasks.all().update(state=State.Failed)
        self.assertEqual(import1337.state, State.Failed)

        # Setup Git objects
        diff1 = MagicMock(change_type='M', a_path='path1a', b_path='path1b')
        diff2 = MagicMock(change_type='D', a_path='path2a', b_path='path2b')

        commit_42, commit_1337, commit_cafe = \
            MagicMock(hexsha='42'), MagicMock(hexsha='1337'), MagicMock(hexsha='cafe')
        commit_42.diff.return_value = [diff1, diff2]

        repo = MagicMock()
        repo.commit.return_value = commit_cafe
        repo.iter_commits.return_value = [commit_42, commit_1337, commit_cafe]

        # Run GitFlow's diff for cafe
        import_cafe = DataImport(
            corpus=self.corpus,
            creator=self.user,
            revision=rev_cafe,
            payload={
                "repo_id": str(self.repo.id),
                "sha": rev_cafe.hash,
            },
            mode=DataImportMode.Repository,
        )
        flow = GitFlow(import_cafe, self.working_dir)
        flow.repo = repo
        diffs = flow.diff()

        self.assertEqual(repo.commit.call_count, 1)
        self.assertEqual(repo.commit.call_args, call('cafe'))
        self.assertEqual(repo.iter_commits.call_count, 1)
        self.assertEqual(repo.iter_commits.call_args, call('cafe'))
        self.assertEqual(commit_42.diff.call_count, 1)
        self.assertEqual(commit_42.diff.call_args, call(commit_cafe))
        self.assertListEqual(diffs, [
            SimpleDiff(DiffType.Modification, 'path1a', 'path1b'),
            SimpleDiff(DiffType.Deletion, 'path2a', 'path2b'),
        ])

    @patch('arkindex.dataimport.git.FileType.get')
    def test_dispatch(self, filetype_mock):
        """
        Test a normal GitFlow file types dispatch
        """
        elt = Element.objects.get(name='Volume 1, page 1r')
        page = Page.objects.get(name='Volume 1, page 1v')
        diff1 = SimpleDiff(DiffType.Modification, 'path1a', 'path1b')
        diff2 = SimpleDiff(DiffType.Modification, 'path2a', 'path2b')

        surface_mock = MagicMock(spec=SurfaceFileType)
        iiif_mock = MagicMock(spec=IIIFFileType)
        surface_mock.diff = diff1
        iiif_mock.diff = diff2
        surface_mock.changed_elements = [elt, ]
        iiif_mock.changed_elements = [page, ]
        filetype_mock.side_effect = [surface_mock, iiif_mock]

        flow = GitFlow(self.dataimport, self.working_dir)
        flow.config = MagicMock()
        changed_elements = flow.dispatch_imports([diff1, diff2])

        self.assertListEqual(changed_elements, [page, elt])

        self.assertEqual(filetype_mock.call_count, 2)
        self.assertListEqual(filetype_mock.call_args_list, [
            call(flow, diff1),
            call(flow, diff2),
        ])
        self.assertEqual(surface_mock.handle.call_count, 1)
        self.assertEqual(surface_mock.handle.call_args, call())
        self.assertEqual(iiif_mock.handle.call_count, 1)
        self.assertEqual(iiif_mock.handle.call_args, call())

    @patch('arkindex.dataimport.git.logger')
    @patch('arkindex.dataimport.git.FileType.get')
    def test_dispatch_errors(self, filetype_mock, logger_mock):
        """
        Test GitFlow handles FileType errors and prints logs
        """
        diffs = [
            SimpleDiff(DiffType.Modification, 'path1a', 'path1b'),
            SimpleDiff(DiffType.Addition, 'path2a', 'path2b'),
        ]
        flow = GitFlow(self.dataimport, self.working_dir)
        flow.config = MagicMock()
        filetype_mock.side_effect = [ValueError('Error 1'), AttributeError('Error 2')]
        changed_pages = flow.dispatch_imports(diffs)

        self.assertListEqual(changed_pages, [])
        self.assertEqual(logger_mock.info.call_count, 1)
        self.assertEqual(logger_mock.info.call_args, call('Fetching file types...'))
        self.assertEqual(logger_mock.debug.call_count, 2)
        self.assertListEqual(logger_mock.debug.call_args_list, [
            call("Parsing diff M from 'path1a' to 'path1b'"),
            call("Parsing diff A from 'path2a' to 'path2b'"),
        ])
        self.assertEqual(logger_mock.warning.call_count, 2)
        self.assertListEqual(logger_mock.warning.call_args_list, [
            call("Error while parsing diff M from 'path1a' to 'path1b': Error 1"),
            call("Error while parsing diff A from 'path2a' to 'path2b': Error 2"),
        ])

    @patch('arkindex.dataimport.git.shutil.rmtree')
    def test_cleanup(self, rmtree_mock):
        """
        Test GitFlow cleans up the repository
        """
        self.assertFalse(rmtree_mock.called)
        GitFlow(self.dataimport, self.working_dir).cleanup()
        self.assertEqual(rmtree_mock.call_count, 1)
        self.assertEqual(rmtree_mock.call_args, call(os.path.join(self.working_dir, 'repo')))
