from unittest.mock import patch, MagicMock
from django.test import TestCase, override_settings
from arkindex_common.ml_tool import MLToolType
from arkindex.dataimport.serializers.ml_tool import MLToolTaskSerializer


@override_settings(ML_CLASSIFIERS_DIR='/nowhere')
class TestMLToolTaskSerializer(TestCase):

    @classmethod
    def _ml_get_fake(cls, directory, tool_type, slug):
        if (tool_type, slug) == (MLToolType.Classifier, 'dummy'):
            return cls.classifier_mock
        if (tool_type, slug) == (MLToolType.Recognizer, 'muddy'):
            return cls.recognizer_mock
        raise ValueError('Missing {.name} {!r}'.format(tool_type, slug))

    @classmethod
    def setUpClass(cls):
        cls.classifier_mock, cls.recognizer_mock = MagicMock(), MagicMock()
        cls.classifier_mock.type, cls.recognizer_mock.type = MLToolType.Classifier, MLToolType.Recognizer
        cls.classifier_mock.slug, cls.recognizer_mock.slug = 'dummy', 'muddy'
        cls.classifier_mock.__str__.return_value = 'Dummy classifier'
        cls.recognizer_mock.__str__.return_value = 'Muddy recognizer'

        cls.ml_get_patch = patch('arkindex.dataimport.serializers.ml_tool.MLTool.get')
        cls.ml_get_mock = cls.ml_get_patch.start()
        cls.ml_get_mock.side_effect = cls._ml_get_fake

    @classmethod
    def tearDownClass(cls):
        cls.ml_get_patch.stop()

    def test_normal(self):
        serializer = MLToolTaskSerializer(data=[
            {'type': 'classifier', 'slug': 'dummy'},
            {'type': 'recognizer', 'slug': 'muddy'},
        ], many=True)
        self.assertTrue(serializer.is_valid(raise_exception=True))
        self.assertEqual(len(serializer.validated_data), 2)
        self.assertListEqual(serializer.validated_data, [
            self.classifier_mock,
            self.recognizer_mock,
        ])

    def test_missing(self):
        serializer = MLToolTaskSerializer(data={
            'type': 'classifier',
            'slug': 'yummy',
        })
        self.assertFalse(serializer.is_valid())
        self.assertEqual(serializer.errors, {
            'non_field_errors': ["Missing Classifier 'yummy'"],
        })

    def test_with_parent(self):
        """
        Test tools can have dependencies on other tools
        """
        serializer = MLToolTaskSerializer(data=[
            {
                'type': 'classifier',
                'slug': 'dummy',
            },
            {
                'type': 'recognizer',
                'slug': 'muddy',
                'parents': [
                    {
                        'type': 'classifier',
                        'slug': 'dummy',
                    }
                ],
            },
        ], many=True)
        self.assertTrue(serializer.is_valid(raise_exception=True))
        self.assertEqual(len(serializer.validated_data), 2)
        self.assertListEqual(serializer.validated_data, [
            self.classifier_mock,
            self.recognizer_mock,
        ])
        self.assertListEqual(self.recognizer_mock.parents, [self.classifier_mock, ])

    def test_duplicate(self):
        """
        Test there can only be one of each tool
        """
        serializer = MLToolTaskSerializer(data=[
            {
                'type': 'classifier',
                'slug': 'dummy',
            },
            {
                'type': 'classifier',
                'slug': 'dummy',
            },
        ], many=True)
        self.assertFalse(serializer.is_valid())
        self.assertDictEqual(serializer.errors, {
            'non_field_errors': ['Tool Dummy classifier is duplicated'],
        })

    def test_missing_dependency(self):
        """
        Tool dependencies should be declared as other tools
        """
        serializer = MLToolTaskSerializer(data=[
            {
                'type': 'recognizer',
                'slug': 'muddy',
                'parents': [
                    {
                        'type': 'classifier',
                        'slug': 'yummy',
                    }
                ],
            },
        ], many=True)
        self.assertFalse(serializer.is_valid())
        self.assertListEqual(serializer.errors, [
            {'parents': [
                {'non_field_errors': ["Missing Classifier 'yummy'"]},
            ]},
        ])

    def test_self_dependent(self):
        """
        Tools should not depend on themselves
        """
        serializer = MLToolTaskSerializer(data={
            'type': 'recognizer',
            'slug': 'muddy',
            'parents': [
                {
                    'type': 'recognizer',
                    'slug': 'muddy',
                }
            ],
        })
        self.assertFalse(serializer.is_valid())
        self.assertDictEqual(serializer.errors, {
            'parents': ['Tool Muddy recognizer depends on itself'],
        })
