from django.contrib import admin
from django import forms
from django.urls import path, reverse
from django.utils.html import format_html
from django_admin_hstore_widget.forms import HStoreFormField
from arkindex.documents.models import \
    Corpus, Page, Element, ElementType, Act, Transcription, MetaData, MLClass, Classification, \
    DataSource, Entity, EntityRole, EntityLink, Region, RegionElement
from arkindex.documents.views import DumpActs
from arkindex.dataimport.models import Event
from enumfields.admin import EnumFieldListFilter


class CorpusAdmin(admin.ModelAdmin):
    list_display = ('id', 'name', 'public', )


class DataSourceAdmin(admin.ModelAdmin):
    list_display = ('id', 'type', 'slug', 'revision', 'internal')
    list_filter = [('type', EnumFieldListFilter), 'internal']
    readonly_fields = ('id', )


class EventInline(admin.TabularInline):
    model = Event


class ClassificationInline(admin.TabularInline):
    model = Classification


class PageAdmin(admin.ModelAdmin):
    list_display = ('id', 'name', 'page_type', 'nb', 'direction', )
    list_filter = [('page_type', EnumFieldListFilter), 'corpus']
    fields = ('id', 'name', 'folio', 'page_type', 'nb', 'direction')
    readonly_fields = ('id', )
    search_fields = ('name', )
    inlines = (EventInline, ClassificationInline, )


class MetaDataAdmin(admin.ModelAdmin):
    list_display = ('id', 'type', 'revision')
    readonly_fields = ('id', 'revision')
    raw_id_fields = ('element', 'entity', )


class MetaDataInline(admin.TabularInline):
    model = MetaData
    raw_id_fields = ('entity', )


class ElementAdmin(admin.ModelAdmin):
    list_display = ('id', 'name', 'type', 'corpus', 'element_actions')
    list_filter = [('type', EnumFieldListFilter), 'corpus']
    fields = ('id', 'type', 'name', 'corpus')
    readonly_fields = ('id', 'element_actions')
    search_fields = ('name', )
    inlines = (MetaDataInline, EventInline)

    def get_urls(self):
        urls = super().get_urls()
        custom_urls = [
            path(
                '<uuid:pk>/acts/',
                self.admin_site.admin_view(DumpActs.as_view()),
                name='dump-acts',
            ),
        ]
        return custom_urls + urls

    def element_actions(self, obj):
        return format_html(
            '<a class="button" href="{}">Dump acts</a>',
            reverse('admin:dump-acts', args=[obj.pk]),
        ) if obj.type == ElementType.Volume else ''
    element_actions.short_description = 'Actions'
    element_actions.allow_tags = True


class ActAdmin(admin.ModelAdmin):
    list_display = ('id', 'name')
    fields = ('id', 'name', 'folio', 'number')
    readonly_fields = ('id', )
    search_fields = ('name', )
    inlines = (MetaDataInline, EventInline)


class TranscriptionAdmin(admin.ModelAdmin):
    list_display = ('id', 'text', 'score', 'element', )
    list_filter = [('type', EnumFieldListFilter), 'source']
    fields = ('id', 'text', 'score', 'element', 'zone', 'source', )
    readonly_fields = ('id', )
    raw_id_fields = ('element', 'zone', )


class MLClassAdmin(admin.ModelAdmin):
    list_display = ('id', 'name', 'corpus')
    list_filter = ('corpus',)
    fields = ('name', 'corpus')


class EntityMetaForm(forms.ModelForm):
    metas = HStoreFormField()


class EntityLinkInLine(admin.TabularInline):
    model = EntityLink
    fk_name = 'parent'
    raw_id_fields = ('child', )


class EntityAdmin(admin.ModelAdmin):
    list_display = ('id', 'name', 'type')
    list_filter = [('type', EnumFieldListFilter), 'corpus']
    readonly_fields = ('id', )
    search_fields = ('name', )
    inlines = (EntityLinkInLine, )
    form = EntityMetaForm


class EntityRoleAdmin(admin.ModelAdmin):
    list_display = ('id', 'parent_name', 'child_name')
    list_filter = ['corpus']
    readonly_fields = ('id', )


class RegionElementInline(admin.TabularInline):
    model = RegionElement
    fields = ('region', 'element', 'ordering')
    raw_id_fields = ('region', 'element')


class RegionAdmin(admin.ModelAdmin):
    list_display = ('id', 'ml_class', 'source')
    readonly_fields = ('id', )
    raw_id_fields = ('zone', )
    inlines = (RegionElementInline, )


admin.site.register(Corpus, CorpusAdmin)
admin.site.register(DataSource, DataSourceAdmin)
admin.site.register(Page, PageAdmin)
admin.site.register(Element, ElementAdmin)
admin.site.register(Act, ActAdmin)
admin.site.register(Transcription, TranscriptionAdmin)
admin.site.register(MLClass, MLClassAdmin)
admin.site.register(MetaData, MetaDataAdmin)
admin.site.register(Entity, EntityAdmin)
admin.site.register(EntityRole, EntityRoleAdmin)
admin.site.register(Region, RegionAdmin)
