from django.utils.decorators import method_decorator
from django.views.decorators.cache import cache_page
from rest_framework.generics import RetrieveAPIView
from rest_framework.exceptions import PermissionDenied
from arkindex_common.enums import TranscriptionType
from arkindex.documents.models import Element, ElementType, Page, Act, Corpus
from arkindex.documents.serializers.iiif import \
    VolumeManifestSerializer, ActManifestSerializer, \
    PageAnnotationListSerializer, PageActAnnotationListSerializer, \
    SurfaceAnnotationListSerializer, TranscriptionSearchAnnotationListSerializer
from arkindex.documents.search import search_transcriptions_filter_post
from arkindex.project.elastic import ESTranscription
from arkindex.project.mixins import SearchAPIMixin


class VolumeManifest(RetrieveAPIView):
    """
    Get a IIIF manifest for a specific volume
    """

    serializer_class = VolumeManifestSerializer

    def get_queryset(self):
        return Element.objects.filter(
            type=ElementType.Volume,
            corpus__in=Corpus.objects.readable(self.request.user),
        )

    @method_decorator(cache_page(3600))
    def get(self, *args, **kwargs):
        return super().get(*args, **kwargs)


class ActManifest(RetrieveAPIView):
    """
    Get a IIIF manfiest for a specific act
    """

    serializer_class = ActManifestSerializer
    queryset = Act.objects.all()

    def get_queryset(self):
        return Act.objects.filter(corpus__in=Corpus.objects.readable(self.request.user))

    @method_decorator(cache_page(3600))
    def get(self, *args, **kwargs):
        return super().get(*args, **kwargs)


class PageAnnotationList(RetrieveAPIView):
    """
    Get a IIIF annotation list for transcriptions of a specific page
    """

    serializer_class = PageAnnotationListSerializer

    def get_queryset(self):
        return Page.objects.filter(corpus__in=Corpus.objects.readable(self.request.user))

    @method_decorator(cache_page(3600))
    def get(self, *args, **kwargs):
        return super().get(*args, **kwargs)


class PageActAnnotationList(RetrieveAPIView):
    """
    Get a IIIF annotation list for acts of a specific page
    """

    serializer_class = PageActAnnotationListSerializer

    def get_queryset(self):
        return Page.objects.filter(corpus__in=Corpus.objects.readable(self.request.user))

    @method_decorator(cache_page(3600))
    def get(self, *args, **kwargs):
        return super().get(*args, **kwargs)


class SurfaceAnnotationList(RetrieveAPIView):
    """
    Get a IIIF annotation list for a single surface
    """
    serializer_class = SurfaceAnnotationListSerializer

    def get_queryset(self):
        return Element.objects.filter(
            type=ElementType.Surface,
            corpus__in=Corpus.objects.readable(self.request.user),
        )

    @method_decorator(cache_page(3600))
    def get(self, *args, **kwargs):
        return super().get(*args, **kwargs)


class TranscriptionSearchAnnotationList(SearchAPIMixin, RetrieveAPIView):
    """
    Search for transcriptions inside an element and get a IIIF annotation list
    """

    serializer_class = TranscriptionSearchAnnotationListSerializer
    elt = None

    def get_element(self):
        if not self.elt:
            self.elt = Element.objects.get(id=self.kwargs['pk'])
            if self.elt.corpus not in Corpus.objects.readable(self.request.user):
                raise PermissionDenied
        return self.elt

    def get_object(self):
        return self.get_queryset()

    def get_search(self, query=None, min_score=0.0, **kwargs):
        return ESTranscription.search() \
            .filter('match', corpus=str(self.get_element().corpus_id)) \
            .filter('match', type=TranscriptionType.Word.value) \
            .filter(
                'terms',
                element=list(
                    Page.objects
                        .get_descending(self.get_element().id)
                        .values_list('id', flat=True)
                ),
            ) \
            .filter('range', score={'gte': min_score}) \
            .query('match', text=query)

    def post_process(self, *args):
        return search_transcriptions_filter_post(*args, element_id=self.get_element().id)
