from django.conf import settings
from rest_framework.generics import ListAPIView
from arkindex.documents.serializers.search import PageSearchResultSerializer, ActSearchResultSerializer, \
    EntitySearchResultSerializer, EntitySearchQuerySerializer
from arkindex.documents.search import search_pages_post, search_acts_post, search_entities_post
from arkindex.project.mixins import SearchAPIMixin
from arkindex.project.elastic import ESPage, ESAct, ESEntity
from elasticsearch_dsl.query import Q, Nested, FunctionScore
from elasticsearch_dsl.function import FieldValueFactor


class SearchAPIView(SearchAPIMixin, ListAPIView):
    """
    A base class for ES search views
    """


class NestedSearchAPIView(SearchAPIView):
    es_document = None

    def get_search(self, corpora_ids=None, query=None, type=None, date_lt=None, date_gte=None, min_score=0.0):
        assert self.es_document is not None, 'Missing ES document. Set the `es_document` attribute'
        assert corpora_ids, 'Must filter by corpora'

        search = self.es_document.search() \
            .sort('volumes', '_score') \
            .source(fields=['date_range']) \
            .filter('terms', corpus=corpora_ids)

        if date_lt or date_gte:
            date_range = {'relation': 'intersects'}
            if date_lt:
                date_range['lt'] = date_lt
            if date_gte:
                date_range['gte'] = date_gte

            search = search.filter('range', date_range=date_range)

        nested_query = Q('range', transcriptions__score={'gte': min_score})
        if query:
            nested_query &= Q('simple_query_string', query=query, fields=['transcriptions.text'])
        if type:
            nested_query &= Q('match', transcriptions__type=type.value)

        search = search.query(Nested(
            path='transcriptions',
            inner_hits={'size': settings.ES_INNER_RESULTS_LIMIT},
            score_mode='sum',
            query=FunctionScore(
                query=nested_query,
                functions=[FieldValueFactor(field='transcriptions.score')],
            ),
        ))

        return search


class PageSearch(NestedSearchAPIView):
    """
    Search and list transcriptions inside pages
    """
    serializer_class = PageSearchResultSerializer
    es_document = ESPage

    def post_process(self, *args, **kwargs):
        return search_pages_post(*args)


class ActSearch(NestedSearchAPIView):
    """
    Search for acts containing a specific word
    within a specific period
    """
    serializer_class = ActSearchResultSerializer
    es_document = ESAct

    def post_process(self, *args, **kwargs):
        return search_acts_post(*args)


class EntitySearch(SearchAPIView):
    """
    Search and list entities
    """
    serializer_class = EntitySearchResultSerializer
    query_serializer_class = EntitySearchQuerySerializer

    def get_search(self, query=None, type=None, corpora_ids=None):
        assert corpora_ids, 'Must filter by corpora'

        search = ESEntity.search() \
            .filter('terms', corpus=corpora_ids)

        if query:
            search = search.query('simple_query_string', query=query, fields=['name'])
        if type:
            search = search.query('match', type=type.value)

        return search

    def post_process(self, *args, **kwargs):
        return search_entities_post(*args)
