from arkindex.documents.models import PageType, PageDirection, PageComplement, Page, ElementType, Element
from arkindex.images.models import Image, Zone
import re
import roman
import logging

logger = logging.getLogger(__name__)

REGEX_FOLIO_PAGE = re.compile(r'^0*(\d+)\s*(?:bis|ter|quat)?\s*([RrVv])?')
REGEX_FOLIO_ROMAN = re.compile(r'^0*([MDCLXVI]+)\s*(?:bis|ter|quat)?\s*([RrVv])?')
FOLIO_WORDS = {
    'contre-plat': PageType.ContrePlat,
    'contre-plat inférieur': PageType.ContrePlat,
    'contre-plat supérieur': PageType.ContrePlat,
    'contre-garde': PageType.ContreGarde,
    'contre-garde supérieur': PageType.ContreGarde,
    'plat': PageType.Plat,
    'plat inférieur': PageType.Plat,
    'plat supérieur': PageType.Plat,
    'garde': PageType.Garde,
    'garde supérieur': PageType.Garde,
    'copie insérée': PageType.Copy,
}


def import_page(volume, image, register, folio=None, order=None):
    '''
    Build a page, inside a volume linked to a register and linked to an image via a zone
    '''
    assert isinstance(volume, Element) and volume.type == ElementType.Volume
    assert isinstance(image, Image)

    # Create a single register element holding all the pages
    element, _ = Element.objects.get_or_create(
        type=ElementType.Register,
        name=register,
        corpus=volume.corpus,
    )
    element.add_parent(volume)

    if order is None:
        order = element.get_next_order(ElementType.Page)

    # Get the image zone if it exists
    poly = [(0, 0), (image.width, 0), (image.width, image.height), (0, image.height), (0, 0)]
    zone = image.zones.filter(
        polygon=poly,
        elements__type=ElementType.Page
    )
    # Get the page or build zone and page
    if zone.exists():
        p = next(el.page for el in zone.first().elements.all() if el.type == ElementType.Page)
    else:
        if folio is None:
            folio = str(order + 1)
        page_type, page_nb, page_direction, page_complement = parse_folio(folio)
        p = Page.objects.create(
            folio=folio,
            name="Page {0} du volume {1}".format(folio, volume.name),
            page_type=page_type,
            nb=page_nb,
            direction=page_direction,
            complement=page_complement,
            zone=Zone.objects.create(polygon=poly, image=image),
            corpus=volume.corpus,
        )

    p.add_parent(element, order=order)
    return p


def parse_folio(folio):
    """
    Parse folio to build a DB representation
    Outputs (type, nb, direction, complement)
    """
    if folio is None:
        return (None, None, None, None)

    def _direction(d):
        if d is None:
            return
        if d.lower() in ('r', 'recto'):
            return PageDirection.Recto
        if d.lower() in ('v', 'verso'):
            return PageDirection.Verso

    complement = None
    for item in PageComplement:
        if item.value in folio:
            complement = item
            break

    # Standard page
    # Removing zeros, because this would match even roman numerals
    page_match = REGEX_FOLIO_PAGE.match(folio.lstrip('0'))
    if page_match is not None:
        groups = page_match.groups()
        page = int(groups[0])
        direction = _direction(groups[1])
        return (PageType.Page, page, direction, complement)

    # Roman page
    roman_match = REGEX_FOLIO_ROMAN.match(folio)
    if roman_match is not None:
        groups = roman_match.groups()
        page = roman.fromRoman(groups[0])
        direction = _direction(groups[1])
        return (PageType.Summary, page, direction, complement)

    # Words
    page_type = FOLIO_WORDS.get(folio.lower())

    return (page_type, None, None, complement)
