from django.conf import settings
from elasticsearch import Elasticsearch
from elasticsearch.helpers import bulk as es_bulk
from elasticsearch.exceptions import NotFoundError
from arkindex.project.elastic import ESTranscription, ESAct, ESPage, ESEntity
import logging
import time
import datetime

logger = logging.getLogger(__name__)


class Indexer(object):

    documents = (
        ESTranscription,
        ESAct,
        ESPage,
        ESEntity,
    )

    def __init__(self, hosts=settings.ELASTIC_SEARCH_HOSTS):
        self.elastic = Elasticsearch(hosts)

    def setup(self):
        """
        Create indexes in ElasticSearch
        """
        for document in self.documents:
            logger.info('Creating or updating index for {}'.format(document.__name__))
            document.init()

    def drop_index(self, document):
        """
        Try to drop an existing index
        """
        try:
            document._index.delete()
            logger.info("Dropped index for {}".format(document.__name__))
        except NotFoundError:
            logger.info("Could not drop index for {} (does not exist)".format(document.__name__))

    def run_index(self, queryset, bulk_size=400):
        """
        Reindex elements from a QuerySet chunk by chunk and output status
        """
        count = queryset.count()
        logger.info("{} items to reindex".format(count))

        start_time, last, indexed = time.time(), 0, 0
        for i in range(0, count, bulk_size):
            percent = int(100.0 * i / count)
            if last != percent or bulk_size > 100:
                secs = time.time() - start_time
                per_second = int(i / secs)
                logger.info(
                    '{0: >3}% => {1: >5} / {2} (indexed: {3}, execution time: {4}, {5} items/s)'.format(
                        percent, i, count, indexed, datetime.timedelta(seconds=secs), per_second))
                last = percent

            indexed += self.index(queryset[i:i+bulk_size])

    def index(self, queryset, retries=3, chunk_size=100, timeout=15):
        """
        Insert or update items into ElasticSearch
        """
        assert retries > 0
        es_doc = queryset.model.es_document

        # Build raw ElasticSearch insert
        actions = [
            es_doc.from_model(item).to_dict(include_meta=True, skip_empty=False)
            for item in queryset.all()
        ]

        # Run actions in bulk
        try:
            nb_insert, _ = es_bulk(
                self.elastic,
                actions,
                chunk_size=chunk_size,
                initial_backoff=2,
                max_retries=2,
                request_timeout=timeout,
                stats_only=True,
                _source=False,
            )
        except Exception as e:
            if retries <= 1:
                logger.warning('Failed to bulk insert into ES : {}'.format(e))
                return 0
            logger.warning('Failed to bulk insert into ES - retrying : {}'.format(e))
            return self.index(
                queryset,
                retries=retries-1,
                chunk_size=max(chunk_size // 2, 1),
                timeout=timeout*2,
            )

        return nb_insert
