#!/usr/bin/env python3
from django.core.management.base import CommandError
from django.conf import settings
from ponos.management.base import PonosCommand
from arkindex.project.argparse import CorpusArgument, ElementArgument
from arkindex.documents.models import Element, ElementType


class Command(PonosCommand):
    help = '(Re)generate thumbnails for elements'
    docker_image = settings.ARKINDEX_APP_IMAGE
    base_recipe = settings.PONOS_RECIPE

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument(
            '--all',
            help='Create thumbnails for every volume in every corpus',
            action='store_true',
            default=False,
        )
        parser.add_argument(
            '--corpus',
            help='ID or part of the name of the corpus to fetch volumes from',
            type=CorpusArgument(),
        )
        parser.add_argument(
            '--element',
            help='ID or part of the name of a single element to build a thumbnail for',
            type=ElementArgument(),
        )
        parser.add_argument(
            '--force',
            help='Recreate thumbnails even if they already exist',
            action='store_true',
            default=False,
        )

    def validate_args(self, corpus=None, element=None, all=False, force=False, **options):
        if all:
            if corpus or element:
                raise CommandError('--all cannot be used together with --corpus or --element')
            return {'elements': Element.objects.filter(type=ElementType.Volume), 'force': force}

        if not corpus:
            raise CommandError('--corpus is required when not using --all')

        if element:
            if not corpus.elements.filter(pk=element.pk).exists():
                raise CommandError('Element {} is not in corpus {}'.format(element, corpus))
            elts = [element, ]
        else:
            elts = corpus.elements.filter(type=ElementType.Volume)

        return {'elements': elts, 'force': force}

    def run(self, elements=[], force=False):
        # TODO: Create a workflow instead when Ponos initial artifacts are implemented
        for element in elements:
            try:
                element.generate_thumbnail(force=force)
            except Exception as e:
                self.stdout.write(self.style.ERROR('Failed to generate thumbnail for {}: {}'.format(element.id, e)))
