#!/usr/bin/env python3
from django.core.management.base import BaseCommand
from django.conf import settings
from django.db.models import Count
from django.utils.text import slugify
from arkindex.documents.models import Element, ElementType, Transcription, Corpus
from arkindex.images.models import ImageServer, Image
from arkindex.project.aws import S3FileStatus
from urllib.parse import urljoin
import time
import requests


class Command(BaseCommand):
    help = 'Display statistics for a Telegraf agent'

    def add_arguments(self, parser):
        parser.add_argument(
            '--influxdb',
            help='Post statistics to the InfluxDB API at a given URL',
            nargs='?',
            const=settings.INFLUXDB_API_URL,
        )

    def handle(self, *args, **options):
        self.options = options

        # Total transcriptions
        self.output(
            'transcriptions',
            total=Transcription.objects.count(),
        )

        # Elements with transcriptions
        self.output(
            'pages',
            total=Element.objects.filter(type=ElementType.Page).count(),
            with_transcriptions=Transcription.objects.values_list('element', flat=True).distinct().count(),
        )

        corpus_total = Corpus.objects.filter(elements__type=ElementType.Volume).annotate(Count('elements'))
        corpus_with_transcriptions = Corpus.objects.raw(
            """
            select v.corpus_id as id, count(distinct v.id) as volumes_count
            from documents_element as e
            inner join (
                select distinct element_id
                from documents_transcription
            ) as t on (t.element_id = e.id)
            inner join documents_elementpath as p on (t.element_id=p.element_id)
            inner join documents_element as v on (v.id=p.path[1])
            group by v.corpus_id;
            """,
        )
        vol_with_transcriptions_count = {
            cwt.id: cwt.volumes_count
            for cwt in corpus_with_transcriptions
        }

        for corpus in corpus_total:
            self.output(
                'corpus',
                tags={'name': slugify(corpus.name)},
                total=corpus.elements__count,
                with_transcriptions=vol_with_transcriptions_count.get(corpus.id, 0),
            )

        # Image server statistics, in three SQL queries
        checked_counts = dict(
            Image.objects
                 .filter(status=S3FileStatus.Checked)
                 .values_list('server')
                 .annotate(Count('server'))
        )
        error_counts = dict(
            Image.objects
                 .filter(status=S3FileStatus.Error)
                 .values_list('server')
                 .annotate(Count('server'))
        )
        for server in ImageServer.objects.annotate(img_total=Count('images')):
            self.output(
                'images',
                tags={'imageserver': server.grafana_tag},
                total=server.img_total,
                checked=checked_counts.get(server.id, 0),
                errors=error_counts.get(server.id, 0),
            )

    def dict_to_str(self, data):
        return ','.join([
            '{}={}'.format(k, str(v).replace(' ', '_'))
            for (k, v) in data.items()
        ])

    def output(self, name, tags={}, **values):
        '''
        Output on stdout in Influx line protocol
        https://docs.influxdata.com/influxdb/v1.6/write_protocols/line_protocol_tutorial/
        '''
        timestamp = time.time() * 10 ** 9  # in nano seconds
        tags_str = ',' + self.dict_to_str(tags) if tags else ''
        values_str = self.dict_to_str(values)
        line = '{}{} {} {:.0f}'.format(name, tags_str, values_str, timestamp)
        self.stdout.write(line)

        if not self.options.get('influxdb'):
            return
        resp = requests.post(urljoin(self.options['influxdb'], 'write'), params={'db': 'arkindex'}, data=line)
        resp.raise_for_status()
