from django.utils.functional import cached_property
from arkindex_common.ml_tool import MLToolType
from arkindex_common.pagexml import PageXmlPage
from arkindex.project.polygon import Polygon
from arkindex_common.enums import TranscriptionType
from arkindex.documents.models import (
    DataSource, ElementType, Page,
    Entity, EntityRole, EntityLink, TranscriptionEntity
)
from arkindex.documents.indexer import Indexer
import functools
import string
import Levenshtein
import os.path
import yaml
import logging

logger = logging.getLogger(__name__)


def best_levenshtein(value, tests):
    score = sorted([
        (i, Levenshtein.ratio(value, i))
        for i in tests
    ], key=lambda x: x[1], reverse=True)
    return score[0]


def load_yaml_configuration(path):
    assert os.path.exists(path)
    with open(path) as f:
        return yaml.safe_load(f)


class PageXmlParser(object):

    def __init__(self, path_or_xml):
        self.pagexml_page = PageXmlPage(path_or_xml)

    @cached_property
    def source(self):
        ds, _ = DataSource.objects.get_or_create(
            type=MLToolType.Recognizer,
            slug='transkribus',
            revision='2013-07-15',
            internal=False,
        )
        return ds

    def get_zone(self, region, page):
        poly = Polygon(region.points)
        z = page.zone.image.zones.filter(polygon=poly).first()
        if not z:
            z = page.zone.image.zones.create(polygon=poly)
        return z

    def create_surface(self, region, page):
        if region.points is None:
            logger.warning('No points in region {}'.format(region.id))
            return None, False
        surface, created = self.get_zone(region, page).elements.get_or_create(
            type=ElementType.Surface,
            corpus=page.corpus,
            defaults={
                'name': 'Surface {}'.format(region.id),
            },
        )
        surface.add_parent(page)
        return surface, created

    def save_surfaces(self, page):
        if not len(self.pagexml_page.page.text_regions):
            logger.warning('No surfaces to save')
            return 0, 0
        region_count, created_count = 0, 0
        for region in self.pagexml_page.page.sort_regions(self.pagexml_page.page.text_regions):
            region_count += 1
            _, created = self.create_surface(region, page)
            created_count += created
        return region_count, created_count

    def create_transcription(self, region, page, type=TranscriptionType.Line):
        if region.text is None:
            logger.warning('No text in region {}'.format(region.id))
            return None, False

        if region.points is None:
            logger.warning('No points in region {}'.format(region.id))
            return None, False

        return self.get_zone(region, page).transcriptions.get_or_create(
            type=type,
            element=page,
            text=region.text,
            source=self.source,
            # Default to 1 if no score is given in the PageXML file
            score=1 if region.confidence is None else region.confidence,
        )

    def index(self, page):
        """
        Create or Update ElasticSearch index for transcriptions and related page
        """
        assert isinstance(page, Page), 'Page should be an Arkindex page'
        es_indexer = Indexer()
        es_indexer.run_index(
            Page.objects.filter(id=page.id),
        )
        es_indexer.run_index(
            page.transcriptions.all(),
        )

    def save_transcriptions(self, page):
        assert isinstance(page, Page), 'Page should be an Arkindex page'
        if self.pagexml_page.page.text_regions is None or not len(self.pagexml_page.page.text_regions):
            logger.warning('No transcriptions to save')
            return []

        region_count, line_count, region_ts_count, line_ts_count = 0, 0, 0, 0
        transcriptions = []
        for region in self.pagexml_page.page.text_regions:
            if len(region.points) < 2:
                logger.warning('Ignoring region {} (not enough points in polygon)'.format(region.id))
                continue

            region_count += 1
            paragraph, created = self.create_transcription(region, page, type=TranscriptionType.Paragraph)
            region_ts_count += created
            if paragraph:
                transcriptions.append((paragraph.id, region))

            for line in region.lines:
                if len(line.points) < 2:
                    logger.warning('Ignoring line {} (not enough points in polygon)'.format(line.id))
                    continue

                line_count += 1
                _, created = self.create_transcription(line, page)
                line_ts_count += created

        logger.info('Parsed {} regions and {} lines and created {} paragraph and {} line transcriptions'.format(
            region_count,
            line_count,
            region_ts_count,
            line_ts_count,
        ))
        self.index(page)
        return transcriptions

    def merge(self, blocks):
        """
        Merge all tags and return blocks
            Regroup tag that continues
            Regroup all first name with a last name
            Create completed value
        """
        def all_tags(block):
            for line in block['lines']:
                for tag in line['tags']:
                    yield tag

        MERGE_FUNCTION = [
            self.regroup,
            self.first_and_last_name,
            self.cesure
        ]
        all_blocks = {}
        for block in blocks:
            index_block = block['reading_order']
            # Filter tags because we can have :
            #   '_prenom' with 'continued'
            #   and 'textStyle' after
            #   and '_prenom' with 'continued' in the next line
            #       => this broke groupby 'continued'
            tags = list(filter(lambda tag: tag['name'].startswith('_'), all_tags(block)))
            for method in MERGE_FUNCTION:
                tags = method(tags)
            all_blocks[index_block] = tags
        return all_blocks

    def regroup(self, items):
        """
        Regroup tag that continues
        """
        def _clean(accu, item):
            assert isinstance(accu, list)

            if accu:
                last = accu[-1]
                # name is type (_prenom, _nom, ...)
                # last and item continued
                # ------------- have the same name
                # item is one line after last
                if (last and 'continued' in last.keys() and 'continued' in item.keys()
                        and last['name'] == item['name'] and last['reading_order']+1 == item['reading_order']):
                    accu[-1]['value'] = None
                    accu[-1]['length'] = 0
                    accu[-1]['reading_order'] = item['reading_order']
                    accu[-1]['items'].append(item)
                    return accu

            new_item = dict(item)
            new_item['items'] = []
            new_item['items'].append(item)
            return accu + [new_item]

        return functools.reduce(_clean, items, [])

    def first_and_last_name(self, tags):
        """
        Regroup all first name with a last name
        """
        def add_used(tags):
            all_tags = []
            for tag in tags:
                all_tags.append(
                    {
                        'used': False,
                        'tag': tag
                    }
                )
            return all_tags

        def distance(first_name, last_name):
            if first_name['offset_in_region'] < last_name['offset_in_region']:
                lentgh = first_name['length']
            else:
                lentgh = last_name['length']
            return abs(int(first_name['offset_in_region']) - int(last_name['offset_in_region'])) - int(lentgh)

        def find_last_name(first_name, list_last_name):
            score = sorted([
                (last_name, distance(first_name, last_name['tag']))
                for last_name in list_last_name
            ], key=lambda x: x[1])
            return score[0][0]

        all_tags = []
        last_names = add_used(filter(lambda tag: tag['name'] == '_nom', tags))

        for tag in tags:
            if tag['name'] == '_prenom' and last_names:
                last_name = find_last_name(tag, last_names)
                tag['items'] += last_name['tag']['items']
                all_tags.append(tag)
                last_name['used'] = True

            elif tag['name'] != '_nom':
                all_tags.append(tag)

        all_tags += [tag['tag'] for tag in last_names if not tag['used']]

        return all_tags

    def cesure(self, tags):
        """
        Create completed value
        """
        for tag in tags:
            value = ''
            length = 0
            for item in tag['items']:
                item['value'] = item['value'].strip()
                if len(item['value']) < 1:
                    continue
                if value.endswith('-') or item['value'].startswith('-'):
                    value = value.rstrip(string.whitespace + '-')
                    item['value'] = item['value'].lstrip(string.whitespace + '-')
                    sep = '-' if item['value'][0].isupper() else ''
                else:
                    sep = ' ' if item['value'][0].isupper() else ''
                value = '{}{}{}'.format(
                    value,
                    sep,
                    item['value']
                ).strip()
                # +1 to count line break after this item
                length += len(item['value']) + 1
            tag['value'] = value
            # -1 to remove the line break after the last item
            tag['length'] = length - 1
        return tags

    def create_objects(self, blocks, page, path, types, ratio=0.90):
        """
        Create entity on transcription, role and link
        """
        entities = []
        for tags in blocks.values():
            new_entities, new_tags = self.create_entities(tags, page, types, ratio)
            entities += new_entities

            self.create_link_transcription_entity(new_tags)
            tags = self.create_roles(new_tags, page, path, ratio)
            self.create_links(tags, page, path, ratio)
        return entities

    def create_entities(self, tags, page, types, ratio):
        """
        Create entity according to 'name' in tag
        """
        entities = []
        for tag in tags:
            if tag['name'] in types.keys():
                entities_name = Entity.objects.filter(
                    corpus__elements=page,
                    type=types[tag['name']]
                ).values('name')
                if entities_name:
                    names = [e['name'] for e in entities_name]
                    name = best_levenshtein(
                        tag['value'],
                        names
                    )
                    score = name[1]
                else:
                    score = 0

                if score < ratio:
                    entity = Entity.objects.create(
                        name=tag['value'],
                        type=types[tag['name']],
                        corpus_id=page.corpus.id,
                        source_id=self.source.id
                    )
                    entities.append(entity.id)
                else:
                    entity = Entity.objects.distinct('name').get(
                        corpus__elements=page,
                        name=name[0],
                        type=types[tag['name']],
                        corpus_id=page.corpus.id
                    )
                tag['entity_id'] = entity.id
        return entities, tags

    def create_link_transcription_entity(self, tags):

        def _clean(accu, item):
            assert isinstance(accu, list)

            if accu:
                last = accu[-1]
                # last item follows current item
                if (last['offset_in_region'] + int(last['length']) + 1 == item['offset_in_region']):
                    last['value'] = '{} {}'.format(
                        last['value'],
                        item['value']
                    )
                    last['length'] = len(last['value'])
                    return accu

            return accu + [item]

        for tag in tags:
            tag['items'] = sorted(tag['items'], key=lambda tag: tag['offset_in_region'])
            if'entity_id' in tag.keys() and 'transcription_id' in tag.keys():
                grouped_items = functools.reduce(_clean, tag['items'], [])
                for item in grouped_items:
                    TranscriptionEntity.objects.get_or_create(
                        entity_id=tag['entity_id'],
                        transcription_id=tag['transcription_id'],
                        offset=item['offset_in_region'],
                        length=item['length']
                    )

    def create_roles(self, tags, page, path, ratio):
        """
        Create role according to '_role' in tag
        """
        roles = load_yaml_configuration(path)
        for tag in tags:
            if '_role' in tag.keys():
                # Entity with parent role
                role = best_levenshtein(
                    tag['_role'],
                    roles.keys()
                )
                if role[1] > ratio:
                    name = role[0]
                    role, created = EntityRole.objects.get_or_create(
                        parent_name=roles[name]['parent_name'],
                        child_name=roles[name]['child_name'],
                        parent_type=roles[name]['parent_type'],
                        child_type=roles[name]['child_type'],
                        corpus_id=page.corpus.id
                    )
                    tag['role_id'] = role.id
                else:
                    logger.warning(
                        'Role {} not recognized for a parent role'.format(tag['_role'])
                    )

        return tags

    def create_links(self, tags, page, path, ratio):
        """
        Create link between two entities with a role
        """
        links = load_yaml_configuration(path)
        for tag in tags:
            if 'role_id' in tag.keys():
                link = best_levenshtein(
                    tag['_role'],
                    links.keys()
                )
                if link[1] > ratio:
                    child = next(
                        filter(lambda tag: '_role' in tag.keys() and tag['_role'] == links[link[0]]['link_with'], tags),
                        False
                    )
                    if child:
                        link, created = EntityLink.objects.get_or_create(
                            parent_id=tag['entity_id'],
                            child_id=child['entity_id'],
                            role_id=tag['role_id']
                        )
                    else:
                        logger.warning(
                            'Child not found for role {}'.format(
                                tag['_role']
                            )
                        )
                else:
                    logger.warning(
                        'Role {} not recognized for a parent role'.format(tag['_role'])
                    )
        return tags

    def create_blocks(self, transcriptions):
        """
        Create blocks with lines, and line with tags
        Order tags according to 'offset'
        Order lines and blocks according to 'reading_order'
        """

        def update_tag(tag, line, index, transcription_id):
            if 'offset' in tag.keys():
                # Save the text value
                tag['value'] = line.text[
                    int(tag['offset']):int(tag['length'])+int(tag['offset'])
                ] if line.text else ''
            tag['reading_order'] = index
            tag['transcription_id'] = transcription_id
            return tag

        blocks = []
        for transcription_id, region in transcriptions:
            # Search the tag 'readingOrder' to select the index
            tag_reading_order = next(filter(lambda tag: tag.name == 'readingOrder', region.tags))
            index_region = int(tag_reading_order.index)
            block = {
                'reading_order': index_region,
                'lines': []
            }

            for line in region.lines:
                # Search the tag 'readingOrder' to select the index
                tag_reading_order = next(filter(lambda tag: tag.name == 'readingOrder', line.tags))
                index_line = int(tag_reading_order.index)
                current_line = {
                    'reading_order': index_line,
                    'tags': [],
                    'length': len(line.text) if line.text else 0
                }
                # Add all tags
                tags = [
                    update_tag(tag.as_dict(), line, index_line, transcription_id)
                    for tag in line.tags
                ]
                current_line['tags'] = tags

                block['lines'].append(current_line)
                # Order tags with their offset
                current_line['tags'] = sorted(
                    current_line['tags'],
                    key=lambda i: int(i['offset']) if 'offset' in i.keys() else -1
                )

            # Order lines with their reading order
            block['lines'] = sorted(
                block['lines'],
                key=lambda i: i['reading_order']
            )
            blocks.append(block)

        # Order blocks with their reading order
        blocks = sorted(blocks, key=lambda i: i['reading_order'])

        # Add an offset in region
        for block in blocks:
            length_region = 0
            for line in block['lines']:
                for tag in line['tags']:
                    tag['offset_in_region'] = length_region
                    tag['offset_in_region'] += int(tag['offset']) if 'offset' in tag.keys() else 0
                # +1 to add line break
                length_region += line['length'] + 1

        return blocks
