from arkindex.documents.models import Corpus, ElementType, Element, Page
from arkindex.documents.pagexml import PageXmlParser
from arkindex.documents.importer import parse_folio
from pathlib import Path
import re
import logging

logger = logging.getLogger(__name__)

REGEX_FOLIO = re.compile(r'(?:fol\.?\s*)?([0-9]+)\s*(bis)?\s*(r|v)?', re.IGNORECASE)


class SurfaceImporter(object):

    def __init__(self, xmlpath=None, basepath=None, corpus=None, **kwargs):
        xmlpath = Path(xmlpath)
        assert isinstance(corpus, Corpus)
        assert xmlpath.exists()
        if basepath:
            basepath = Path(basepath)
            assert basepath.is_dir(), str(basepath)
            self.basepath = basepath
        else:
            self.basepath = xmlpath
        self.corpus = corpus
        self.xmlpath = xmlpath
        self.volumes = {}

    def get_xml_paths(self):
        """
        Get (volume name, path) tuples for XML files
        """
        if self.xmlpath.is_file():
            path = self.xmlpath.relative_to(self.basepath)
            yield path.parts[-2], path
            return

        for path in self.xmlpath.glob('**/*.xml'):
            path = path.relative_to(self.basepath)
            if len(path.parts) <= 1:  # File is in root folder
                logger.warning('File {} is not in a subfolder'.format(str(path)))
                continue
            yield path.parts[0], path

    def find_volume(self, name):
        """
        Get an Element instance from a volume name by case-insensitive match in the corpus
        """
        if name in self.volumes:
            return self.volumes[name]
        try:
            self.volumes[name] = self.corpus.elements.get(type=ElementType.Volume, name__iexact=name)
            logger.debug('Matched volume "{}" for volume name "{}"'.format(self.volumes[name], name))
            return self.volumes[name]
        except Element.DoesNotExist:
            logger.warning('No matching volume found for "{}"'.format(name))
        except Element.MultipleObjectsReturned:
            logger.warning('Multiple matching volumes found for "{}"'.format(name))

    def get_xml_pages(self):
        """
        Get (Page, path) tuples for XML files
        """
        for volume_name, path in self.get_xml_paths():
            volume = self.find_volume(volume_name)
            if not volume:
                continue
            page_type, nb, direction, complement = parse_folio(path.stem)
            try:
                page = Page.objects \
                    .get_descending(volume.id) \
                    .prefetch_related('zone__image') \
                    .get(
                        page_type=page_type,
                        nb=nb,
                        direction=direction,
                        complement=complement,
                    )
                yield page, path
            except ValueError:
                logger.warning("Could not parse folio '{}' in volume '{}'".format(path.stem, volume_name))
            except Page.DoesNotExist:
                logger.warning("No page found for folio '{}' in volume '{}'".format(path.stem, volume_name))
            except Page.MultipleObjectsReturned:
                logger.warning(
                    "Multiple matching pages found for folio '{}' in volume '{}'".format(path.stem, volume_name)
                )

    def run(self, dry=False):
        xml_count, surfaces_count, created_surfaces_count = 0, 0, 0

        for page, path in self.get_xml_pages():
            xml_count += 1
            parser = PageXmlParser(self.basepath / path)

            if dry:  # Handle dry run
                logger.info("{}\t{}".format(path, page.zone.image.path))
                continue

            region_count, created_count = parser.save_surfaces(page)
            surfaces_count += region_count
            created_surfaces_count += created_count

        logger.info("Parsed {} and created {} surfaces from {} XML files".format(
            surfaces_count, created_surfaces_count, xml_count,
        ))
