from arkindex.documents.models import \
    Element, ElementType, Act, Page, PageType, PageDirection, PageComplement, Corpus
from arkindex.documents.indexer import Indexer
from django.db import transaction
from collections import namedtuple, OrderedDict
from pathlib import Path
import csv
import datetime
import logging
import re
import time

logger = logging.getLogger(__name__)

REGEX_FOLIO = re.compile(r'([0-9]+)\s*(bis|ter|quat)?\s*(r|v)?', re.IGNORECASE)


class ParsedFolio(object):

    def __init__(self, number, complement, direction):
        assert isinstance(number, int) or not number
        assert isinstance(complement, PageComplement) or not complement
        assert isinstance(direction, PageDirection) or not direction
        self.number = number
        self.complement = complement
        self.direction = direction

    def __str__(self):
        return ''.join([
            str(self.number),
            self.complement.value if self.complement is not None else '',
            self.direction.value[0],
        ])

    def __eq__(self, other):
        return other and (self.number, self.complement, self.direction) == \
            (other.number, other.complement, other.direction)

    def __hash__(self):
        return hash((self.number, self.complement, self.direction))

    @staticmethod
    def from_page(page):
        return ParsedFolio(page.nb, page.complement, page.direction)

    @staticmethod
    def from_raw(number, complement, direction):
        try:
            pagecplt = PageComplement(complement)
        except ValueError:
            pagecplt = None
        pagedir = PageDirection.Verso \
            if direction and direction.startswith('v') \
            else PageDirection.Recto
        return ParsedFolio(int(number), pagecplt, pagedir)


def parse_folios(text, pages=None):
    """
    Get a list of ParsedFolio (number, complement, direction) from a folio string
    If `pages` is set, will check the parsed folios against the pages and return intermediate folios.
    `pages` is a list of Page instances
    """
    text = text.strip()
    folios = REGEX_FOLIO.findall(text)
    if len(folios) < 1:
        raise ValueError('Improperly formatted folio string')
    results = [
        ParsedFolio.from_raw(*folio)
        for folio in folios
    ]
    if pages is None or len(results) < 2:
        return results

    return get_folios_interval(pages, results[0], results[-1])


def get_folios_interval(pages, first_folio, last_folio):
    """
    Fetch all intermediate folios in a list of pages
    """
    assert all(isinstance(p, Page) for p in pages), 'Page instances required'

    # Get first and last pages from list
    try:
        first_page, last_page = [next(
            p for p in pages
            if p.nb == f.number and p.complement == f.complement and p.direction == f.direction
        ) for f in (first_folio, last_folio)]
    except StopIteration:
        raise ValueError('No associated page found')

    # Check folios aren't in reverse order
    if last_page < first_page:
        raise ValueError("Folio seems to be in reverse order")

    # Rebuild folios to add intermediate folios
    return [
        ParsedFolio.from_page(page)
        for page in sorted(pages)
        if first_page <= page <= last_page
    ]


ActRaw = namedtuple('ActRaw', 'number, folio')


class CorpusSurfaceLinker(object):
    """
    A wrapper to SurfaceLinker, to perform linking on many volumes at once.
    Faster to write than rewriting the SurfaceLinker to handle multiple volumes
    """

    def __init__(self, csvpath, corpus, col_number=0, col_folio=1):
        self.csvpath = Path(csvpath)
        assert self.csvpath.exists()
        assert isinstance(corpus, Corpus)
        self.corpus = corpus
        self.col_number = col_number
        self.col_folio = col_folio
        self.volumes = {}

    def find_csv(self):
        """
        Look for all CSV files and return (volume name, CSV reader) tuples
        """
        if self.csvpath.is_file():
            with self.csvpath.open() as f:
                yield (self.csvpath.stem, csv.reader(f, delimiter=';'))
            return

        for path in self.csvpath.glob('**/*.csv'):
            with path.open() as f:
                yield (path.stem, csv.reader(f, delimiter=';'))

    def find_volume(self, name):
        """
        Get an Element instance from a volume name by case-insensitive match in the corpus
        """
        if name in self.volumes:
            return self.volumes[name]
        try:
            self.volumes[name] = self.corpus.elements.get(type=ElementType.Volume, name__iexact=name)
            logger.debug('Matched volume "{}" for volume name "{}"'.format(self.volumes[name], name))
            return self.volumes[name]
        except Element.DoesNotExist:
            logger.warning('No matching volume found for "{}"'.format(name))
        except Element.MultipleObjectsReturned:
            logger.warning('Multiple matching volumes found for "{}"'.format(name))

    def run(self, dry=False):
        """
        Start SurfaceLinkers from CSV files
        """
        for volume_name, reader in self.find_csv():
            volume = self.find_volume(volume_name)
            if not volume:
                continue

            next(reader)  # Skip header line
            rows = [
                (
                    row[self.col_number],
                    row[self.col_folio],
                )
                for row in reader
                if row and len(row[0]) > 0
            ]

            linker = SurfaceLinker(volume=volume, data=rows)
            linker.run(dry=dry)


class SurfaceLinker(object):
    """
    Creates acts and links surfaces from a CSV file for a specific volume
    """

    def __init__(self, volume, data):
        """
        data: list of (act number, folio text) tuples
        """
        assert isinstance(volume, Element)
        assert volume.type == ElementType.Volume
        assert isinstance(data, (list, tuple)), 'Data must be a matrix.'
        assert all(isinstance(row, (list, tuple)) for row in data), 'Data must be a matrix.'
        assert all(len(row) == 2 for row in data), 'All rows must have 2 columns.'
        self.volume = volume

        # Act number is NOT always a number, so we cannot cast to int
        # but we can sort with integers, by removing the string parts (bis & co)
        self.acts_raw = sorted([
            ActRaw(row[0], row[1])
            for row in data
        ], key=lambda a: int(re.sub(r'[^\d+]', '', a.number)))

    def find_pages(self):
        # TODO: support all page types !
        logger.info('Fetching pages')
        self.pages = list(
            Page.objects.get_descending(self.volume.id)
                        .filter(page_type=PageType.Page)
                        .prefetch_related('zone'))

    def find_surfaces(self):
        logger.info('Fetching surfaces')

        # Load all surfaces linked to our pages
        image_ids = [p.zone.image_id for p in self.pages]
        surfaces = Element.objects \
                          .filter(type=ElementType.Surface, zone__image_id__in=image_ids) \
                          .prefetch_related('zone') \
                          .order_by('zone__image_id') \
                          .distinct()
        self.surfaces_count = surfaces.count()
        if surfaces.count() < 1:
            # Warn and continue; will create acts without any surfaces
            logger.warning('No surfaces found in volume {}'.format(self.volume.name))

        # Link a folio with its available surfaces
        # and sort the surface by their median height position
        # so we can process them in the natural reading order
        self.surfaces = {
            ParsedFolio.from_page(page): sorted([
                surface
                for surface in surfaces
                if surface.zone.image_id == page.zone.image_id
            ], key=lambda s: s.zone.polygon.center.y)
            for page in self.pages
        }

    def find_acts(self):
        logger.info('Fetching existing acts')
        self.acts = Act.objects.get_descending(self.volume.id)

    def link_acts_folios(self):
        logger.info('Computing links')
        self.parsed_acts = 0
        self.act_folios = OrderedDict()

        # The only trustable folio is the act starting folio - we start by finding all pages between this act's
        # starting folio and the next act's starting folio. Conflicts will be solved later.
        for index, act in enumerate(self.acts_raw):
            logger.debug('Parsing act {} of volume {} with folio "{}"'.format(
                act.number, self.volume.name, act.folio))

            # First find this act's folios
            try:
                current_folios = parse_folios(act.folio, pages=self.pages)
            except ValueError as e:
                logger.warning('Could not parse folio "{}" (volume {}, act number {}): {}'.format(
                    act.folio, self.volume.name, act.number, str(e)))
                continue

            # If this act isn't the last one, get the next act
            if index + 1 < len(self.acts_raw):
                next_act = self.acts_raw[index+1]

                # Find the next act's folios and pick all folios between
                # our act's first folio and the next act's first folio
                try:
                    next_folios = parse_folios(next_act.folio, pages=self.pages)
                    self.act_folios[act] = get_folios_interval(self.pages, current_folios[0], next_folios[0])
                except ValueError as e:
                    logger.warning('Could not parse folio "{}" (volume {}, act number {}): {}'.format(
                        next_act.folio, self.volume.name, next_act.number, str(e)))
            else:
                self.act_folios[act] = current_folios

    def resolve_conflicts(self):
        logger.info('Resolving conflicts')
        for page in self.pages:
            folio = ParsedFolio.from_page(page)
            # Count all references of this folio found in acts and compare with surfaces
            refcount = sum(folios.count(folio) for folios in self.act_folios.values())
            surfcount = len(self.surfaces[folio])

            while refcount > surfcount:  # Loop until conflict is solved
                # Find the first act that refers to this folio
                act, folios = next((act, folios) for act, folios in self.act_folios.items() if folio in folios)
                # Remove the reference and reassign to act
                folios.remove(folio)
                self.act_folios[act] = folios
                refcount -= 1

    def link_surfaces(self):
        self.linked_surfaces = 0
        self.acts_surfaces = {}

        for act, folios in self.act_folios.items():
            if len(folios) < 1:
                logger.warning(
                    'Act {} of volume {} has no associated surfaces'.format(act.number, self.volume.name),
                )
                continue

            # This is a greedy algorithm, so we need to handle different cases
            # in different ways:
            # * when a single folio is requested, only the next available item is picked
            # * when more than a single folio is requested, we pick the last item
            #   on the page to continue through the natural reading order
            start_index = 0 if len(folios) == 1 else -1

            # Take last available surface from the first folio
            found_surfaces = self._pop_surfaces(folios[0], index=start_index)

            # Take all the surfaces from intermediate folios
            if len(folios) > 2:
                for folio in folios[1:-1]:
                    found_surfaces.extend(self._pop_surfaces(folio))

            # Take the first available from the last folio, if there is one
            if len(folios) > 1:
                found_surfaces.extend(self._pop_surfaces(folios[-1], index=0))

            self.acts_surfaces[act.number] = found_surfaces
            self.linked_surfaces += len(found_surfaces)
            self.parsed_acts += 1

    def find_missing_acts(self):
        """
        Act numbers should follow each other; try to create potentially missing acts.
        """
        logger.info('Looking for potentially missing acts')

        for i, act in enumerate(self.acts_raw):
            # Last act
            if i + 1 >= len(self.acts_raw):
                return

            # Get integer without complements
            act_int = int(re.sub(r'[^\d].*$', '', act.number))

            # First act
            if i == 0 and act_int > 1:
                for new_number in range(1, act_int):
                    self.acts_raw.insert(0, ActRaw(str(new_number), ''))
                    logger.warning('Created potentially missing act {}'.format(str(new_number)))
                continue

            next_act = self.acts_raw[i + 1]
            next_act_int = int(re.sub(r'[^\d]', '', next_act.number))

            # If an act seems missing (e.g. 13-15)
            if next_act_int - act_int > 1:
                # Add the act (e.g. 14)
                self.acts_raw.insert(i + 1, ActRaw(str(act_int + 1), ''))
                logger.warning('Created potentially missing act {}'.format(str(act_int + 1)))

    @transaction.atomic
    def save_act(self, act_raw):
        logger.debug('Saving act {} of volume {} with folio "{}"'.format(
            act_raw.number, self.volume.name, act_raw.folio))

        # Get the act
        act = next((a for a in self.acts if a.number == act_raw.number), None)
        if act is None:
            act = Act.objects.create(
                name="Act {}".format(act_raw.number),
                number=act_raw.number,
                folio=act_raw.folio,
                corpus=self.volume.corpus,
            )
            self.created_acts += 1

        # Link act to volume, if not already done
        act.add_parent(self.volume)

        # Link surfaces to volumes
        for surface in self.acts_surfaces.get(act_raw.number, []):
            surface.add_parent(act)

        return act.id

    def save(self):
        logger.info('Saving to database')
        self.created_acts = 0
        act_ids = []
        for act_raw in self.acts_raw:
            act_ids.append(self.save_act(act_raw))

        # Index all affected acts in ElasticSearch
        Indexer().run_index(Act.objects.filter(id__in=act_ids))

    def run(self, dry=False):
        self.start_time = time.time()
        logger.info('{} for volume {} started'.format('Dry run' if dry else 'Linking', self.volume.name))
        self.find_pages()
        self.find_surfaces()
        self.find_acts()
        self.link_acts_folios()
        self.resolve_conflicts()
        self.link_surfaces()
        self.find_missing_acts()
        self.save()
        self.end_time = time.time()
        self.print_stats(dry_run=dry)

    def print_stats(self, dry_run=False):
        logger.info('Ran in {}'.format(datetime.timedelta(seconds=self.end_time - self.start_time)))
        logger.info('Parsed {} acts'.format(self.parsed_acts))
        logger.info('Found {} surfaces'.format(self.surfaces_count))
        logger.info('{} {} surfaces to acts'.format(dry_run and 'Would link' or 'Linked', self.linked_surfaces))
        if not dry_run:
            logger.info('Created {} acts'.format(self.created_acts))

    def _pop_surfaces(self, folio, index=None):
        """
        Find all available surfaces in a folio and remove them from the available surfaces.
        If `index` is set, will pick one surface at this index.
        Returns a list of found surfaces.
        Will log warnings if folio does not exist or has no surfaces available.
        """
        if folio not in self.surfaces:
            logger.warning('Folio {} not found in surfaces of volume {}'.format(
                folio, self.volume.name))
            return []

        if not self.surfaces[folio]:
            logger.warning('Ran out of surfaces for folio {} of volume {}'.format(
                folio, self.volume.name))
            return []

        if index is not None:
            return [self.surfaces[folio].pop(index)]
        else:
            # Use full list of surfaces
            s = self.surfaces[folio]
            self.surfaces[folio] = []
            return s
