from unittest.mock import patch, call
from django.core.management import call_command
from arkindex_common.enums import MetaType
from arkindex.project.tests import FixtureTestCase
from arkindex.project.elastic import ESTranscription, ESAct, ESPage, ESEntity
from arkindex.documents.models import Element, ElementType, Act, Page, Transcription, Entity, EntityType, \
    DataSource, MLToolType


class TestReindexCommand(FixtureTestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        source = DataSource.objects.create(type=MLToolType.NER, slug='entity', internal=True)
        cls.indexer_patch = patch('arkindex.documents.management.commands.reindex.Indexer')
        cls.vol = Element.objects.get(type=ElementType.Volume, name="Volume 1")
        cls.entity = cls.corpus.entities.create(type=EntityType.Misc, name='Dummy entity', source=source)
        page = Page.objects.get(name='Volume 1, page 1r')
        page.metadatas.create(name='Dummy metadata', value='Dummy', type=MetaType.Text, entity=cls.entity)

    def setUp(self):
        super().setUp()
        self.indexer_mock = self.indexer_patch.start()

    def tearDown(self):
        super().tearDown()
        self.indexer_patch.stop()

    def _assert_all_acts(self, call_args):
        """
        Helper method to assert run_index is called to reindex all acts
        Required because of self.assertQuerysetEqual, the only way Django has to compare querysets in tests
        """
        (queryset, ), kwargs = call_args
        self.assertQuerysetEqual(
            queryset,
            map(repr, Act.objects.all()),
            ordered=False,
        )
        self.assertDictEqual(kwargs, {'bulk_size': 100})

    def _assert_all_entities(self, call_args):
        """
        Helper method to assert run_index is called to reindex all entities
        """
        (queryset, ), kwargs = call_args
        self.assertQuerysetEqual(
            queryset,
            map(repr, Entity.objects.all()),
            ordered=False,
        )
        self.assertDictEqual(kwargs, {'bulk_size': 400})

    def _assert_all_pages(self, call_args):
        """
        Helper method to assert run_index is called to reindex all pages
        """
        (queryset, ), kwargs = call_args
        self.assertQuerysetEqual(
            queryset,
            map(repr, Page.objects.filter(
                id__in=Transcription.objects.values_list('element', flat=True).distinct(),
            )),
            ordered=False,
        )
        self.assertDictEqual(kwargs, {'bulk_size': 100})

    def _assert_all_transcriptions(self, call_args):
        """
        Helper method to assert run_index is called to reindex all transcriptions
        """
        (queryset, ), kwargs = call_args
        self.assertQuerysetEqual(
            queryset,
            map(repr, Transcription.objects.all()),
            ordered=False,
        )
        self.assertDictEqual(kwargs, {'bulk_size': 400})

    def _assert_all(self):
        """
        Helper method to assert run_index is called three times to reindex everything
        """
        self.assertEqual(self.indexer_mock().run_index.call_count, 4)
        acts_call, entities_call, pages_call, ts_call = sorted(self.indexer_mock().run_index.call_args_list, key=repr)
        self._assert_all_acts(acts_call)
        self._assert_all_entities(entities_call)
        self._assert_all_transcriptions(ts_call)
        self._assert_all_pages(pages_call)

    def _assert_volume_acts(self, call_args):
        """
        Helper method to assert run_index is called to reindex all acts in a volume
        """
        (queryset, ), kwargs = call_args
        self.assertQuerysetEqual(
            queryset,
            map(repr, Act.objects.get_descending(self.vol.id)),
            ordered=False,
        )
        self.assertDictEqual(kwargs, {'bulk_size': 100})

    def _assert_volume_entities(self, call_args):
        """
        Helper method to assert run_index is called to reindex all entities in a volume
        """
        (queryset, ), kwargs = call_args
        self.assertQuerysetEqual(
            queryset,
            map(repr, Entity.objects.filter(
                metadatas__element__in=Element.objects.get_descending(self.vol.id))
                ),
            ordered=False,
        )
        self.assertDictEqual(kwargs, {'bulk_size': 400})

    def _assert_volume_pages(self, call_args):
        """
        Helper method to assert run_index is called to reindex all pages in a volume
        """
        (queryset, ), kwargs = call_args
        self.assertQuerysetEqual(
            queryset,
            map(repr, Page.objects.get_descending(self.vol.id)),
            ordered=False,
        )
        self.assertDictEqual(kwargs, {'bulk_size': 100})

    def _assert_volume_transcriptions(self, call_args):
        """
        Helper method to assert run_index is called to reindex all transcriptions in a volume
        """
        (queryset, ), kwargs = call_args
        self.assertQuerysetEqual(
            queryset,
            map(repr, Transcription.objects.filter(
                element__in=Element.objects.get_descending(self.vol.id),
            )),
            ordered=False,
        )
        self.assertDictEqual(kwargs, {'bulk_size': 400})

    def _assert_volume(self):
        """
        Helper method to assert run_index is called three times to reindex a volume
        """
        self.assertEqual(self.indexer_mock().run_index.call_count, 4)
        acts_call, entities_call, pages_call, ts_call = sorted(self.indexer_mock().run_index.call_args_list, key=repr)
        self._assert_volume_acts(acts_call)
        self._assert_volume_entities(entities_call)
        self._assert_volume_pages(pages_call)
        self._assert_volume_transcriptions(ts_call)

    def _assert_corpus_acts(self, call_args):
        """
        Helper method to assert run_index is called to reindex all acts in a corpus
        """
        (queryset, ), kwargs = call_args
        self.assertQuerysetEqual(
            queryset,
            map(repr, Act.objects.filter(corpus=self.corpus)),
            ordered=False,
        )
        self.assertDictEqual(kwargs, {'bulk_size': 100})

    def _assert_corpus_entities(self, call_args):
        """
        Helper method to assert run_index is called to reindex all entities in a corpus
        """
        (queryset, ), kwargs = call_args
        self.assertQuerysetEqual(
            queryset,
            map(repr, Entity.objects.filter(corpus=self.corpus)),
            ordered=False,
        )
        self.assertDictEqual(kwargs, {'bulk_size': 400})

    def _assert_corpus_pages(self, call_args):
        """
        Helper method to assert run_index is called to reindex all pages in a corpus
        """
        (queryset, ), kwargs = call_args
        self.assertQuerysetEqual(
            queryset,
            map(repr, Page.objects.filter(corpus=self.corpus)),
            ordered=False,
        )
        self.assertDictEqual(kwargs, {'bulk_size': 100})

    def _assert_corpus_transcriptions(self, call_args):
        """
        Helper method to assert run_index is called to reindex all transcriptions in a corpus
        """
        (queryset, ), kwargs = call_args
        self.assertQuerysetEqual(
            queryset,
            map(repr, Transcription.objects.filter(element__corpus=self.corpus)),
            ordered=False,
        )
        self.assertDictEqual(kwargs, {'bulk_size': 400})

    def _assert_corpus(self):
        """
        Helper method to assert run_index is called three times to reindex a volume
        """
        self.assertEqual(self.indexer_mock().run_index.call_count, 4)
        acts_call, entities_call, pages_call, ts_call = sorted(self.indexer_mock().run_index.call_args_list, key=repr)
        self._assert_corpus_acts(acts_call)
        self._assert_corpus_entities(entities_call)
        self._assert_corpus_pages(pages_call)
        self._assert_corpus_transcriptions(ts_call)

    def test_acts(self):
        """
        Test the reindex command can reindex acts
        """
        call_command(
            'reindex',
            acts=True,
        )
        self.assertEqual(self.indexer_mock().drop_index.call_count, 0)
        self.assertEqual(self.indexer_mock().setup.call_count, 0)
        self.assertEqual(self.indexer_mock().run_index.call_count, 1)
        self._assert_all_acts(self.indexer_mock().run_index.call_args)

    def test_entities(self):
        """
        Test the reindex command can reindex entities
        """
        call_command(
            'reindex',
            entities=True,
        )
        self.assertEqual(self.indexer_mock().drop_index.call_count, 0)
        self.assertEqual(self.indexer_mock().setup.call_count, 0)
        self.assertEqual(self.indexer_mock().run_index.call_count, 1)
        self._assert_all_entities(self.indexer_mock().run_index.call_args)

    def test_pages(self):
        """
        Test the reindex command can reindex pages
        """
        call_command(
            'reindex',
            pages=True,
        )
        self.assertEqual(self.indexer_mock().drop_index.call_count, 0)
        self.assertEqual(self.indexer_mock().setup.call_count, 0)
        self.assertEqual(self.indexer_mock().run_index.call_count, 1)
        self._assert_all_pages(self.indexer_mock().run_index.call_args)

    def test_transcriptions(self):
        """
        Test the reindex command can reindex transcriptions
        """
        call_command(
            'reindex',
            transcriptions=True,
        )
        self.assertEqual(self.indexer_mock().drop_index.call_count, 0)
        self.assertEqual(self.indexer_mock().setup.call_count, 0)
        self.assertEqual(self.indexer_mock().run_index.call_count, 1)
        self._assert_all_transcriptions(self.indexer_mock().run_index.call_args)

    def test_all(self):
        """
        Test the reindex command reindexes everything by default
        """
        call_command(
            'reindex',
        )
        self.assertEqual(self.indexer_mock().drop_index.call_count, 0)
        self.assertEqual(self.indexer_mock().setup.call_count, 0)
        self._assert_all()

    def test_volume(self):
        """
        Test the reindex command can restrict indexing to a specific volume
        """
        call_command(
            'reindex',
            volume=self.vol,
        )
        self.assertEqual(self.indexer_mock().drop_index.call_count, 0)
        self.assertEqual(self.indexer_mock().setup.call_count, 0)
        self._assert_volume()

    def test_corpus(self):
        """
        Test the reindex command can restrict indexing to a specific corpus
        """
        call_command(
            'reindex',
            corpus=self.corpus,
        )
        self.assertEqual(self.indexer_mock().drop_index.call_count, 0)
        self.assertEqual(self.indexer_mock().setup.call_count, 0)
        self._assert_corpus()

    def test_drop(self):
        """
        Test the reindex command can drop and recreate indexes
        """
        call_command(
            'reindex',
            drop=True,
        )
        self.assertEqual(self.indexer_mock().drop_index.call_count, 4)
        self.assertCountEqual(self.indexer_mock().drop_index.call_args_list, [
            call(ESTranscription),
            call(ESAct),
            call(ESPage),
            call(ESEntity),
        ])
        self.assertEqual(self.indexer_mock().setup.call_count, 1)
        self._assert_all()
