from unittest.mock import patch
from django.urls import reverse
from rest_framework import status
from arkindex_common.enums import MetaType
from arkindex.project.tests import FixtureAPITestCase
from arkindex.project.elastic import ESAct
from arkindex.documents.models import Element, Act, ElementType, MetaData
from arkindex.documents.dates import DateType, InterpretedDate
from arkindex.images.models import Image, Zone


class TestAct(FixtureAPITestCase):

    def setUp(self):
        # Create an act with some surfaces
        self.img = Image.objects.get(path='img6', server=self.imgsrv)
        self.act = Act.objects.create(
            corpus=self.corpus,
            name='sister act',
            number='1',
        )

        self.surfaces = []
        for i in range(3):
            x, y = i*100, i*100
            z = Zone.objects.create(polygon=[(x, y), (x+50, y), (x+50, y+50), (x, y+50), (x, y)], image=self.img)
            s = Element.objects.create(
                name='surface {}'.format(i),
                corpus=self.corpus,
                type=ElementType.Surface,
                zone=z,
            )
            s.add_parent(self.act)
            self.surfaces.append(s)

    def test_act_retrieve(self):
        '''
        Test simple GET query
        '''
        response = self.client.get(reverse('api:act-edit', kwargs={'pk': self.act.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['id'], str(self.act.id))
        self.assertEqual(data['name'], 'sister act')
        self.assertEqual(data['number'], '1')
        self.assertEqual(data['parents'], [])
        self.assertEqual(data['metadata'], [])
        self.assertEqual(len(data['children']), 3)
        for i, child in enumerate(data['children']):
            self.assertEqual(child['type'], 'surface')
            self.assertEqual(child['corpus']['id'], str(self.corpus.id))
            self.assertEqual(child['corpus']['name'], self.corpus.name)
            self.assertEqual(child['name'], self.surfaces[i].name)
            self.assertEqual(child['id'], str(self.surfaces[i].id))

    def test_act_metadatas(self):
        self.assertFalse(MetaData.objects.exists())

        # No metadatas in api
        response = self.client.get(reverse('api:act-edit', kwargs={'pk': self.act.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(response.json()['metadata'], [])

        # Create some metadatas
        metas = [
            self.act.metadatas.create(type=MetaType.Text, name='test', value='aha'),
            self.act.metadatas.create(type=MetaType.Date, name='origin', value='2010/01'),
            self.act.metadatas.create(type=MetaType.Location, name='place', value='somewhere'),
        ]

        # Retrieve them through api, sorted by name
        response = self.client.get(reverse('api:act-edit', kwargs={'pk': self.act.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            response.json()['metadata'],
            [{'id': str(metas[1].id),
              'name': 'origin',
              'type': 'date',
              'value': '2010/01',
              'revision': None,
              'dates': []},
             {'id': str(metas[2].id),
              'name': 'place',
              'type': 'location',
              'value': 'somewhere',
              'revision': None,
              'dates': []},
             {'id': str(metas[0].id),
              'name': 'test',
              'type': 'text',
              'value': 'aha',
              'revision': None,
              'dates': []}]
        )

    @patch('arkindex.documents.models.MetaData.get_dates')
    def test_build_search_index_dates(self, get_dates_mock):
        self.act.metadatas.create(type=MetaType.Date, name='date', value='something')

        # Test Upper bound
        get_dates_mock.return_value = [InterpretedDate(1420, 5, type=DateType.Upper), ]
        date_range = ESAct.from_model(self.act).to_dict().get('date_range')
        self.assertDictEqual(date_range, {'lt': '1420-05||+1M'})

        # Test Lower bound
        get_dates_mock.return_value = [InterpretedDate(1418, type=DateType.Lower), ]
        date_range = ESAct.from_model(self.act).to_dict().get('date_range')
        self.assertDictEqual(date_range, {'gte': '1418'})

        # Test with both Lower and Upper bound
        get_dates_mock.return_value = [
            InterpretedDate(1418, type=DateType.Lower),
            InterpretedDate(1428, type=DateType.Upper),
        ]
        date_range = ESAct.from_model(self.act).to_dict().get('date_range')
        self.assertDictEqual(date_range, {'gte': '1418', 'lt': '1428||+1y'})

        # Test an exact date
        get_dates_mock.return_value = [InterpretedDate(1666, 2, 3), ]
        date_range = ESAct.from_model(self.act).to_dict().get('date_range')
        self.assertDictEqual(date_range, {'gte': '1666-02-03', 'lt': '1666-02-03||+1d'})
