from django.contrib.auth.models import AnonymousUser
from django.urls import reverse
from rest_framework import status

from arkindex.documents.models import Corpus, DataSource, Element, ElementType, MLClass, Page
from arkindex.project.tests import FixtureAPITestCase


class TestCorpus(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.anon = AnonymousUser()

        cls.corpus_public = Corpus.objects.get(name='Unit Tests', public=True)
        cls.corpus_private = Corpus.objects.create(name='B Private')
        cls.user.corpus_right.create(corpus=cls.corpus_private, user=cls.user, can_write=True)
        cls.corpus_hidden = Corpus.objects.create(name='C Hidden')

    def test_anon(self):
        # An anonymous user has only access to public
        response = self.client.get(reverse('api:corpus'))
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertListEqual(
            response.json(),
            [
                {
                    'id': str(self.corpus_public.id),
                    'rights': ['read'],
                    'public': True,
                    'name': 'Unit Tests',
                    'description': '',
                }
            ]
        )

    def test_user(self):
        # A normal user has access to public + its private
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:corpus'))
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertListEqual(
            response.json(),
            [
                {
                    'id': str(self.corpus_private.id),
                    'rights': ['read', 'write'],
                    'public': False,
                    'name': 'B Private',
                    'description': '',
                },
                {
                    'id': str(self.corpus_public.id),
                    'rights': ['read', 'write', 'admin'],
                    'public': True,
                    'name': 'Unit Tests',
                    'description': '',
                }
            ]
        )

    def test_admin(self):
        # An admin has access to everything
        self.client.force_login(self.superuser)
        response = self.client.get(reverse('api:corpus'))
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertListEqual(
            response.json(),
            [
                {
                    'id': str(c.id),
                    'rights': ['read', 'write', 'admin'],
                    'public': c.public,
                    'name': c.name,
                    'description': '',
                }
                for c in Corpus.objects.all().order_by('name')
            ]
        )

    def test_mixin(self):
        vol1 = Element.objects.get(name="Volume 1", type=ElementType.Volume)
        vol2 = Element.objects.get(name="Volume 2", type=ElementType.Volume)

        response = self.client.get('{}?corpus={}'.format(reverse('api:elements'), str(self.corpus_private.id)))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        response = self.client.get('{}?corpus={}'.format(reverse('api:elements'), str(self.corpus_public.id)))
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        data = response.json()
        self.assertEqual(len(data['results']), 2)
        self.assertCountEqual(
            [r['id'] for r in data['results']],
            [str(vol1.id), str(vol2.id)],
        )

    def test_create(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:corpus'), {
            'name': 'New Corpus',
            'description': 'Some description',
            'public': False,
        })
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        corpus = Corpus.objects.get(name='New Corpus', description='Some description', public=False)
        right = corpus.corpus_right.get(user=self.user)
        self.assertTrue(right.can_write)
        self.assertTrue(right.can_admin)

    def test_create_requires_login(self):
        response = self.client.post(reverse('api:corpus'), {
            'name': 'New Corpus',
            'description': 'Some description',
            'public': False,
        })
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_retrieve_public(self):
        response = self.client.get(reverse('api:corpus-retrieve', kwargs={'pk': self.corpus_public.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            'id': str(self.corpus_public.id),
            'name': self.corpus_public.name,
            'description': self.corpus_public.description,
            'public': True,
            'rights': ['read'],
        })

    def test_retrieve(self):
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:corpus-retrieve', kwargs={'pk': self.corpus_private.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            'id': str(self.corpus_private.id),
            'name': self.corpus_private.name,
            'description': self.corpus_private.description,
            'public': False,
            'rights': ['read', 'write'],
        })

    def test_retrieve_requires_login(self):
        response = self.client.get(reverse('api:corpus-retrieve', kwargs={'pk': self.corpus_private.id}))
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_update(self):
        self.client.force_login(self.user)
        response = self.client.patch(reverse('api:corpus-retrieve', kwargs={'pk': self.corpus_private.id}), {
            'name': 'new name',
            'description': 'new description',
        })
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.corpus_private.refresh_from_db()
        self.assertEqual(self.corpus_private.name, 'new name')
        self.assertEqual(self.corpus_private.description, 'new description')

    def test_update_requires_login(self):
        response = self.client.patch(reverse('api:corpus-retrieve', kwargs={'pk': self.corpus_private.id}), {
            'name': 'new name',
            'description': 'new description',
        })
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_delete(self):
        self.client.force_login(self.user)
        response = self.client.delete(reverse('api:corpus-retrieve', kwargs={'pk': self.corpus_private.id}))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.assertFalse(Corpus.objects.filter(id=self.corpus_private.id).exists())

    def test_delete_requires_login(self):
        response = self.client.delete(reverse('api:corpus-retrieve', kwargs={'pk': self.corpus_private.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_pages_list(self):
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:corpus-pages', kwargs={'pk': self.corpus_public.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['count'], 6)
        self.assertEqual(len(response.data['results']), 6)
        self.assertIsNone(response.data['next'])
        self.assertIsNone(response.data['previous'])

        # Add a class to filter with: without classification, must not return any result
        text = MLClass.objects.create(name='text', corpus=self.corpus_public)
        response = self.client.get(
            reverse('api:corpus-pages', kwargs={'pk': self.corpus_public.id}), {'ml_class': text.id})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            "count": 0,
            "number": 1,
            "next": None,
            "previous": None,
            "results": []
        })

        # Add a classification: must return one result
        page = Page.objects.get(name='Volume 1, page 1v')
        classification = page.classifications.create(
            source=DataSource.objects.get(slug='test'),
            ml_class=text,
        )
        response = self.client.get(
            reverse('api:corpus-pages', kwargs={'pk': self.corpus_public.id}),
            {'ml_class': text.id},
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            "count": 1,
            "number": 1,
            "next": None,
            "previous": None,
            "results": [{
                'id': str(page.id),
                'page_type': 'page',
                'nb': 1,
                'direction': 'verso',
                'display_name': 'Page no. 1, verso',
                'image': {
                    'id': str(page.zone.image_id),
                    'path': 'img2',
                    's3_url': None,
                    'width': 1000,
                    'height': 1000,
                    'url': 'http://server/img2',
                    'status': 'unchecked'
                },
                'classifications': [
                    {
                        'id': str(classification.id),
                        'source': {
                            'id': str(classification.source.id),
                            'type': 'recognizer',
                            'slug': 'test',
                            'revision': '4.2',
                            'internal': False
                        },
                        'ml_class': {
                            'id': str(text.id),
                            'name': 'text'
                        },
                        'state': 'pending',
                        'confidence': None
                    }
                ],
                'folio': '1v'
                }
            ]
        })
