from unittest.mock import patch, call
from arkindex.project.tests import FixtureTestCase
from arkindex.documents.models import Element, ElementType
from arkindex.documents.surface_link import CorpusSurfaceLinker
import os.path

FIXTURES = os.path.join(
    os.path.dirname(os.path.realpath(__file__)),
    'act_samples',
)


class TestCorpusSurfaceLinker(FixtureTestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.vol1 = Element.objects.get(type=ElementType.Volume, name='Volume 1')
        cls.vol2 = Element.objects.get(type=ElementType.Volume, name='Volume 2')

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.surface_linker_mock = patch('arkindex.documents.surface_link.SurfaceLinker').start()

    def setUp(self):
        self.surface_linker_mock.reset_mock()

    def test_check_directory(self):
        with self.assertRaises(AssertionError):
            CorpusSurfaceLinker('/nowhere', self.corpus)

    def test_find_csv(self):
        csl = CorpusSurfaceLinker(FIXTURES, self.corpus)
        results = [(name, list(reader)) for name, reader in csl.find_csv()]
        self.assertCountEqual(results, [
            ('volume 1', [
                ['Act', 'Folio', 'anything else'],
                ['1', '1r-1v', 'a'],
                ['2', '1v-3r', 'b'],
                ['3', '4-4v', 'c'],
            ]),
            ('volume 2', [
                ['Act', 'Folio', 'something else'],
                ['7', '5r-7v', 'x'],
                ['8', '1-3r', 'y'],
                ['9', '4v-4', 'z'],
            ]),
        ])

    def test_find_volume(self):
        csl = CorpusSurfaceLinker(FIXTURES, self.corpus)
        self.assertEqual(csl.find_volume("volume 1"), self.vol1)
        self.assertIsNone(csl.find_volume("waldo"))
        self.assertDictEqual(csl.volumes, {"volume 1": self.vol1})

    def test_run(self):
        csl = CorpusSurfaceLinker(FIXTURES, self.corpus)
        csl.run()
        self.assertEqual(self.surface_linker_mock.call_count, 2)
        self.assertCountEqual(self.surface_linker_mock.call_args_list, [
            call(volume=self.vol1, data=[
                ('1', '1r-1v'),
                ('2', '1v-3r'),
                ('3', '4-4v'),
            ]),
            call(volume=self.vol2, data=[
                ('7', '5r-7v'),
                ('8', '1-3r'),
                ('9', '4v-4'),
            ]),
        ])
        self.assertEqual(self.surface_linker_mock().run.call_count, 2)
        self.assertEqual(self.surface_linker_mock().run.call_args_list, [call(dry=False), call(dry=False)])

    def test_dry_run(self):
        csl = CorpusSurfaceLinker(FIXTURES, self.corpus)
        csl.run(dry=True)
        self.assertEqual(self.surface_linker_mock.call_count, 2)
        self.assertEqual(self.surface_linker_mock().run.call_count, 2)
        self.assertEqual(self.surface_linker_mock().run.call_args_list, [call(dry=True), call(dry=True)])

    def test_col_numbers(self):
        csl = CorpusSurfaceLinker(FIXTURES, self.corpus, col_folio=0, col_number=2)
        csl.run()
        self.assertEqual(self.surface_linker_mock.call_count, 2)
        self.assertCountEqual(self.surface_linker_mock.call_args_list, [
            call(volume=self.vol1, data=[
                ('a', '1'),
                ('b', '2'),
                ('c', '3'),
            ]),
            call(volume=self.vol2, data=[
                ('x', '7'),
                ('y', '8'),
                ('z', '9'),
            ]),
        ])
        self.assertEqual(self.surface_linker_mock().run.call_count, 2)
        self.assertEqual(self.surface_linker_mock().dry_run.call_count, 0)
