from django.urls import reverse
from unittest.mock import patch
from rest_framework import status
from arkindex_common.enums import MetaType
from arkindex.documents.models import Element, ElementType, TranscriptionType, DataSource, \
    Corpus, Entity, EntityType, EntityRole, EntityLink, TranscriptionEntity, MLToolType
from arkindex.project.tests import FixtureAPITestCase


class TestEntitiesAPI(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.entity_source = DataSource.objects.create(type=MLToolType.NER, slug='entity', internal=True)
        cls.source = DataSource.objects.get(slug='test')
        cls.private_corpus = Corpus.objects.create(name='private')

    def setUp(self):
        super().setUp()
        self.page = Element.objects.filter(type=ElementType.Page, corpus=self.corpus).first().page
        self.entity = Entity.objects.create(
            type=EntityType.Person,
            corpus=self.corpus,
            name='entity 1',
            source_id=self.source.id
        )
        self.entity_bis = Entity.objects.create(
            type=EntityType.Location,
            corpus=self.corpus,
            name='entity 2',
            source_id=self.source.id
        )
        self.role = EntityRole.objects.create(
            parent_name="parent",
            child_name="child",
            parent_type=EntityType.Person,
            child_type=EntityType.Location,
            corpus=self.corpus
        )
        self.transcription = self.page.transcriptions.create(
            text='Some transcribed text',
            source_id=self.source.id,
            type=TranscriptionType.Line
        )
        self.metadata = self.page.metadatas.create(
            name='test 1',
            type=MetaType.Entity,
            value='Blah',
            element=self.page,
            entity=self.entity,
        )
        self.transcriptionentity = TranscriptionEntity.objects.create(
            transcription=self.transcription,
            entity=self.entity_bis,
            offset=2,
            length=len(self.entity_bis.name)
        )
        self.tr_entities_sample = {
            'entity': str(self.entity.id),
            'offset': 4,
            'length': len(self.entity.name)
        }

    def test_get_entity(self):
        response = self.client.get(reverse('api:entity-details', kwargs={'pk': str(self.entity.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['id'], str(self.entity.id))
        self.assertEqual(data['type'], self.entity.type.value)
        self.assertEqual(data['corpus']['id'],  str(self.corpus.id))
        self.assertEqual(data['name'], self.entity.name)

    def test_get_entity_elements(self):
        with self.assertNumQueries(6):
            response = self.client.get(reverse('api:entity-elements', kwargs={'pk': str(self.entity.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        results = data['results']
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0]['id'], str(self.page.id))
        self.assertEqual(results[0]['name'], self.page.name)

    def test_get_entity_elements_corpus_acl(self):
        self.client.force_login(self.user)
        self.page.corpus = self.private_corpus
        self.page.save()
        response = self.client.get(reverse('api:entity-elements', kwargs={'pk': str(self.entity.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(response.json().get('results'), [])

    def test_get_entity_elements_from_transcription(self):
        elt = Element.objects.create(corpus=self.corpus, type=ElementType.Act, name='001')
        elt_tr = elt.transcriptions.create(source_id=self.source.id, text='goodbye')
        TranscriptionEntity.objects.create(transcription=elt_tr, entity=self.entity, offset=42, length=7)
        with self.assertNumQueries(6):
            response = self.client.get(reverse('api:entity-elements', kwargs={'pk': str(self.entity.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        results = data['results']
        self.assertEqual(len(results), 2)
        self.assertListEqual(
            results,
            [{
                'id': str(e.id),
                'type': e.type.value,
                'name': e.name,
                'corpus': {
                    'id': str(self.corpus.id),
                    'name': self.corpus.name,
                    'public': self.corpus.public
                },
                'thumbnail_url': e.thumbnail.s3_url,
                'thumbnail_put_url': e.thumbnail.s3_put_url
            } for e in (elt, self.page)]
        )

    def test_get_role_in_corpus(self):
        response = self.client.get(reverse('api:corpus-roles', kwargs={'pk': str(self.corpus.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        results = data['results']
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0]['parent_type'], self.role.parent_type.value)
        self.assertEqual(results[0]['child_type'], self.role.child_type.value)
        self.assertEqual(results[0]['parent_name'], self.role.parent_name)
        self.assertEqual(results[0]['child_name'], self.role.child_name)

    def test_create_role_in_corpus(self):
        data = {
            'parent_name': 'other parent',
            'child_name': 'other child',
            'parent_type': EntityType.Person.value,
            'child_type': EntityType.Location.value
        }
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:corpus-roles', kwargs={'pk': str(self.corpus.id)}), data=data)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        self.assertIn('id', data)
        role = self.corpus.roles.get(id=data['id'])
        self.assertEqual(role.parent_type, EntityType.Person)
        self.assertEqual(role.child_type, EntityType.Location)
        self.assertEqual(role.parent_name, 'other parent')
        self.assertEqual(role.child_name, 'other child')

    def test_create_role_already_exist(self):
        data = {
            'parent_name': self.role.parent_name,
            'child_name': self.role.child_name,
            'parent_type': self.role.parent_type.value,
            'child_type': self.role.child_type.value,
        }
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:corpus-roles', kwargs={'pk': str(self.corpus.id)}), data=data)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        data = response.json()
        self.assertEqual(data, {
            'corpus': ['Role already exists in this corpus'],
            'id': str(self.corpus.id)
            })

    def test_create_role_not_verified(self):
        data = {
            'parent_name': 'other parent',
            'child_name': 'other child',
            'parent_type': EntityType.Organization.value,
            'child_type': EntityType.Location.value
        }
        response = self.client.post(reverse('api:corpus-roles', kwargs={'pk': str(self.corpus.id)}), data=data)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        data = response.json()
        self.assertEqual(data, {
            'corpus': ['You do not have write access to this corpus'],
            'id': [str(self.corpus.id)]
            })

    @patch('arkindex.project.serializer_fields.MLTool.get')
    def test_create_entity_person(self, ml_get_mock):
        ml_get_mock.return_value.type = self.entity_source.type
        ml_get_mock.return_value.slug = self.entity_source.slug
        ml_get_mock.return_value.version = self.entity_source.revision
        self.entity_source.internal = True
        self.entity_source.save()
        data = {
            'name': 'entity',
            'type': EntityType.Person.value,
            'corpus': str(self.corpus.id),
            'metas': {
                'key': 'value',
                'other key': 'other value'
            },
            'ner': self.entity_source.slug
        }
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:entity-create'), data=data, format='json')
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        entity = Entity.objects.get(id=response.json()['id'])
        self.assertEqual(entity.name, 'entity')
        self.assertEqual(entity.raw_dates, None)
        self.assertEqual(entity.source, self.entity_source)

    @patch('arkindex.project.serializer_fields.MLTool.get')
    def test_create_entity_number(self, ml_get_mock):
        ml_get_mock.return_value.type = self.entity_source.type
        ml_get_mock.return_value.slug = self.entity_source.slug
        ml_get_mock.return_value.version = self.entity_source.revision
        self.entity_source.internal = True
        self.entity_source.save()
        data = {
            'name': '300g',
            'type': EntityType.Number.value,
            'corpus': str(self.corpus.id),
            'metas': {
                'key': 'value',
                'other key': 'other value'
            },
            'ner': self.entity_source.slug
        }
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:entity-create'), data=data, format='json')
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        entity = Entity.objects.get(id=response.json()['id'])
        self.assertEqual(entity.name, '300g')
        self.assertEqual(entity.raw_dates, None)

    @patch('arkindex.project.serializer_fields.MLTool.get')
    def test_create_entity_date(self, ml_get_mock):
        ml_get_mock.return_value.type = self.entity_source.type
        ml_get_mock.return_value.slug = self.entity_source.slug
        ml_get_mock.return_value.version = self.entity_source.revision
        self.entity_source.internal = True
        self.entity_source.save()
        data = {
            'name': '1789',
            'type': EntityType.Date.value,
            'corpus': str(self.corpus.id),
            'metas': {
                'key': 'value',
                'other key': 'other value'
            },
            'ner': self.entity_source.slug
        }
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:entity-create'), data=data, format='json')
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        entity = Entity.objects.get(id=response.json()['id'])
        self.assertEqual(entity.name, '1789')
        self.assertEqual(entity.raw_dates, entity.name)

    def test_create_entity_requires_login(self):
        data = {
            'name': 'entity',
            'type': EntityType.Person.value,
            'corpus': str(self.corpus.id),
            'metas': {
                'key': 'value',
                'other key': 'other value'
            },
            'ner': self.entity_source.slug
        }
        response = self.client.post(reverse('api:entity-create'), data=data, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_create_link(self):
        child = Entity.objects.create(
            type=EntityType.Location,
            corpus=self.corpus,
            name="child",
            source_id=self.source.id
        )
        data = {
            'parent': str(self.entity.id),
            'child': str(child.id),
            'role': str(self.role.id)
        }
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:entity-link-create'), data=data, format='json')
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        link = EntityLink.objects.get(id=response.json()['id'])
        self.assertEqual(link.parent.id, self.entity.id)
        self.assertEqual(link.child.id, child.id)
        self.assertEqual(link.role.id, self.role.id)

    def test_create_link_requires_login(self):
        child = Entity.objects.create(
            type=EntityType.Location,
            corpus=self.corpus,
            name="child",
            source_id=self.source.id
        )
        data = {
            'parent': str(self.entity.id),
            'child': str(child.id),
            'role': str(self.role.id)
        }
        response = self.client.post(reverse('api:entity-create'), data=data, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_create_link_error(self):
        child = Entity.objects.create(
            type=EntityType.Person,
            corpus=self.corpus,
            name="child",
            source_id=self.source.id
        )
        data = {
            'parent': str(self.entity.id),
            'child': str(child.id),
            'role': str(self.role.id)
        }
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:entity-link-create'), data=data, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_create_transcription_entity(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:transcription-entity-create', kwargs={'pk': str(self.transcription.id)}),
            data=self.tr_entities_sample,
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        transcription_entity = TranscriptionEntity.objects.get(
            transcription=self.transcription,
            entity=self.entity,
            offset=self.tr_entities_sample['offset'],
            length=self.tr_entities_sample['length']
        )
        self.assertIsNotNone(transcription_entity)
        self.assertDictEqual(
            response.json(),
            {
                'entity': str(transcription_entity.entity.id),
                'offset': transcription_entity.offset,
                'length': transcription_entity.length
            }
        )

    def test_create_transcription_entity_requires_login(self):
        response = self.client.post(
            reverse('api:transcription-entity-create', kwargs={'pk': str(self.transcription.id)}),
            data=self.tr_entities_sample,
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_create_transcription_entity_wrong_acl(self):
        self.client.force_login(self.user)
        self.page.corpus = self.private_corpus
        self.page.save()
        response = self.client.post(
            reverse('api:transcription-entity-create', kwargs={'pk': str(self.transcription.id)}),
            data=self.tr_entities_sample,
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {'transcription': ['invalid UUID or Corpus write-access is forbidden']}
        )

    def test_create_transcription_entity_different_corpus(self):
        self.client.force_login(self.user)
        ent = Entity.objects.create(
            type=EntityType.Person,
            corpus=self.private_corpus,
            name="a private entity",
            source_id=self.source.id
        )
        self.tr_entities_sample.update({'entity': ent.id})
        response = self.client.post(
            reverse('api:transcription-entity-create', kwargs={'pk': str(self.transcription.id)}),
            data=self.tr_entities_sample,
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {'__all__': ['Transcription and Entity are not in the same corpus']}
        )

    def test_create_transcription_entity_exists(self):
        self.client.force_login(self.user)
        TranscriptionEntity.objects.create(
            transcription=self.transcription,
            entity=self.entity,
            offset=4,
            length=len(self.entity.name)
        )
        response = self.client.post(
            reverse('api:transcription-entity-create', kwargs={'pk': str(self.transcription.id)}),
            data=self.tr_entities_sample,
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_create_transcription_entity_key_missing(self):
        self.client.force_login(self.user)
        data = self.tr_entities_sample
        del data['length']
        response = self.client.post(
            reverse('api:transcription-entity-create', kwargs={'pk': str(self.transcription.id)}),
            data=data,
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_list_transcription_entities_open(self):
        response = self.client.get(reverse('api:transcription-entities', kwargs={'pk': str(self.transcription.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_list_transcription_entities_wrong_acl(self):
        self.transcription.element.corpus = self.private_corpus
        self.transcription.element.save()
        TranscriptionEntity.objects.create(
            transcription=self.transcription,
            entity=self.entity,
            offset=4,
            length=len(self.entity.name)
        )
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:transcription-entities', kwargs={'pk': str(self.transcription.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json()['count'], 0)

    def test_list_transcription_entities(self):
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:transcription-entities', kwargs={'pk': str(self.transcription.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['count'], 1)
        self.assertListEqual(
            data['results'],
            [{
                'entity': {
                    'id': str(self.entity_bis.id),
                    'name': self.entity_bis.name,
                    'type': self.entity_bis.type.value,
                    'metas': None,
                    'validated': self.entity_bis.validated,
                    'dates': [],
                    'source': {
                        'id': str(self.source.id),
                        'type': self.source.type.value,
                        'slug': self.source.slug,
                        'revision': self.source.revision,
                        'internal': self.source.internal,
                    }
                },
                'length': self.transcriptionentity.length,
                'offset': self.transcriptionentity.offset
            }]
        )

    def test_list_element_entities(self):
        """
        List an element's entities sorted depending on their origin (metadata or transcriptions)
        """
        md = self.page.metadatas.create(name='some_metadata', type=MetaType.Entity, value='something')
        md.entity = self.entity_bis
        md.save()
        with self.assertNumQueries(9):
            response = self.client.get(reverse('api:element-entities', kwargs={'pk': str(self.page.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertDictEqual(
            data,
            {
                'id': str(self.page.id),
                'type': self.page.type.value,
                'name': self.page.name,
                'transcriptions': [{
                    'entity': {
                        'id': str(t.entity.id),
                        'name': t.entity.name,
                        'type': t.entity.type.value,
                        'metas': None,
                        'validated': t.entity.validated,
                        'dates': [],
                        'source': {
                            'id': str(self.source.id),
                            'type': self.source.type.value,
                            'slug': self.source.slug,
                            'revision': self.source.revision,
                            'internal': self.source.internal,
                        }
                    },
                    'offset': t.offset,
                    'length': t.length
                } for t in TranscriptionEntity.objects.filter(transcription=self.transcription)],
                'metadata': [{
                    'entity': {
                        'id': str(m.entity.id),
                        'name': m.entity.name,
                        'type': m.entity.type.value,
                        'metas': None,
                        'validated': m.entity.validated,
                        'dates': [],
                        'source': {
                            'id': str(self.source.id),
                            'type': self.source.type.value,
                            'slug': self.source.slug,
                            'revision': self.source.revision,
                            'internal': self.source.internal,
                        }
                    },
                    'id': str(m.id),
                    'type': m.type.value,
                    'name': m.name,
                    'value': m.value,
                    'revision': m.revision,
                    'dates': []
                } for m in self.page.metadatas.all()]
            }
        )

    def test_list_element_entities_wrong_acl(self):
        self.page.corpus = self.private_corpus
        self.page.save()
        response = self.client.get(reverse('api:element-entities', kwargs={'pk': str(self.page.id)}))
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_delete_entity(self):
        self.client.force_login(self.user)
        response = self.client.delete(reverse('api:entity-details', kwargs={'pk': str(self.entity.id)}))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        with self.assertRaises(Entity.DoesNotExist):
            Entity.objects.get(id=self.entity.id)

    def test_delete_entity_not_verified(self):
        response = self.client.delete(reverse('api:entity-details', kwargs={'pk': str(self.entity_bis.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_validated_entity(self):
        self.client.force_login(self.user)
        response = self.client.patch(
            reverse('api:entity-details', kwargs={'pk': self.entity_bis.id}),
            {'validated': True},
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.entity_bis.refresh_from_db()
        self.assertEqual(self.entity_bis.validated, True)
        self.assertEqual(self.entity_bis.moderator, self.user)

    def test_unvalidated_entity(self):
        self.client.force_login(self.user)
        response = self.client.patch(
            reverse('api:entity-details', kwargs={'pk': self.entity_bis.id}),
            {'validated': False},
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.entity_bis.refresh_from_db()
        self.assertEqual(self.entity_bis.validated, False)
        self.assertEqual(self.entity_bis.moderator, None)

    def test_validated_entity_not_verified(self):
        response = self.client.patch(
            reverse('api:entity-details', kwargs={'pk': self.entity_bis.id}),
            {'validated': True},
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_unvalidated_entity_not_verified(self):
        response = self.client.patch(
            reverse('api:entity-details', kwargs={'pk': self.entity_bis.id}),
            {'validated': False},
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_list_element_links(self):
        link = EntityLink.objects.create(parent=self.entity, child=self.entity_bis, role=self.role)
        response = self.client.get(reverse('api:element-links', kwargs={'pk': str(self.page.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertListEqual(
            data['results'],
            [{
                'id': str(link.id),
                'parent': {
                    'id': str(self.entity.id),
                    'name': self.entity.name,
                    'type': self.entity.type.value,
                    'metas': None,
                    'validated': self.entity.validated,
                    'dates': [],
                    'source': {
                        'id': str(self.source.id),
                        'type': self.source.type.value,
                        'slug': self.source.slug,
                        'revision': self.source.revision,
                        'internal': self.source.internal,
                    }
                },
                'child': {
                    'id': str(self.entity_bis.id),
                    'name': self.entity_bis.name,
                    'type': self.entity_bis.type.value,
                    'metas': None,
                    'validated': self.entity_bis.validated,
                    'dates': [],
                    'source': {
                        'id': str(self.source.id),
                        'type': self.source.type.value,
                        'slug': self.source.slug,
                        'revision': self.source.revision,
                        'internal': self.source.internal,
                    }
                },
                'role': {
                    'id': self.role.id,
                    'parent_name': self.role.parent_name,
                    'child_name': self.role.child_name,
                    'parent_type': self.role.parent_type.value,
                    'child_type': self.role.child_type.value
                }
            }]
        )
