from django.urls import reverse
from rest_framework import status

from arkindex.documents.models import ClassificationState, DataSource, MLClass, Page
from arkindex.project.tests import FixtureAPITestCase


class TestClasses(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.text = MLClass.objects.create(name='text', corpus=cls.corpus)
        cls.page = Page.objects.get(name='Volume 1, page 1v')

    def _create_classification(self):
        return self.page.classifications.create(
            source=DataSource.objects.get(slug='test'),
            ml_class=self.text,
            confidence=.5,
        )

    def test_classification_creation(self):
        """
        Ensure classification creation works and set auto fields correctly
        """
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:classification-create'), {
            'page': str(self.page.id),
            'ml_class': str(self.text.id)
        })
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(response.data['source']['type'], 'classifier')
        self.assertEqual(response.data['source']['slug'], 'manual')
        self.assertDictEqual(response.data['ml_class'], {
            'id': str(self.text.id),
            'name': 'text'
        })
        self.assertEqual(response.data['state'], ClassificationState.Validated.value)
        self.assertEqual(response.data['confidence'], 1)

    def test_classification_creation_without_permission(self):
        """
        Without permission, should not raise a 403 Forbidden: indicates that the resource exists
        Indicates that page PK does not exist instead
        """
        response = self.client.post(reverse('api:classification-create'), {
            'page': str(self.page.id),
            'ml_class': str(self.text.id)
        })
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.data, {
            'page': [
                'Invalid pk "{!s}" - object does not exist.'.format(self.page.id),
            ]
        })

    def test_classification_validate(self):
        self.client.force_login(self.user)
        classification = self._create_classification()
        response = self.client.put(reverse('api:classification-validate', kwargs={'pk': classification.id}))
        self.assertEqual(response.status_code, status.HTTP_202_ACCEPTED)
        self.assertDictEqual(response.json(), {
            'id': str(classification.id),
            'source': {
                'id': str(classification.source.id),
                'type': classification.source.type.value,
                'slug': classification.source.slug,
                'revision': classification.source.revision,
                'internal': classification.source.internal
            },
            'ml_class': {
                'id': str(classification.ml_class.id),
                'name': classification.ml_class.name
            },
            'state': ClassificationState.Validated.value,
            'confidence': classification.confidence
        })

        # Ensure moderator has been set
        classification.refresh_from_db()
        self.assertEqual(classification.moderator, self.user)

    def test_classification_validate_without_permissions(self):
        classification = self._create_classification()
        response = self.client.put(reverse('api:classification-validate', kwargs={'pk': classification.id}))
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_classification_reject(self):
        self.client.force_login(self.user)
        classification = self._create_classification()
        response = self.client.put(reverse('api:classification-reject', kwargs={'pk': classification.id}))
        self.assertEqual(response.status_code, status.HTTP_202_ACCEPTED)
        self.assertDictEqual(response.json(), {
            'id': str(classification.id),
            'source': {
                'id': str(classification.source.id),
                'type': classification.source.type.value,
                'slug': classification.source.slug,
                'revision': classification.source.revision,
                'internal': classification.source.internal
            },
            'ml_class': {
                'id': str(classification.ml_class.id),
                'name': classification.ml_class.name
            },
            'state': ClassificationState.Rejected.value,
            'confidence': classification.confidence
        })

        # Ensure moderator has been set
        classification.refresh_from_db()
        self.assertEqual(classification.moderator, self.user)

    def test_classification_reject_without_permissions(self):
        classification = self._create_classification()
        response = self.client.put(reverse('api:classification-reject', kwargs={'pk': classification.id}))
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_classification_can_still_be_moderated(self):
        self.client.force_login(self.user)
        classification = self._create_classification()
        classification.moderator = self.user
        classification.state = ClassificationState.Validated.value
        classification.save()

        # First try to reject
        response = self.client.put(reverse('api:classification-reject', kwargs={'pk': classification.id}))
        self.assertEqual(response.status_code, status.HTTP_202_ACCEPTED)
        self.assertDictEqual(response.json(), {
            'id': str(classification.id),
            'source': {
                'id': str(classification.source.id),
                'type': classification.source.type.value,
                'slug': classification.source.slug,
                'revision': classification.source.revision,
                'internal': classification.source.internal
            },
            'ml_class': {
                'id': str(classification.ml_class.id),
                'name': classification.ml_class.name
            },
            'state': ClassificationState.Rejected.value,
            'confidence': classification.confidence
        })

        # Then try to validate
        response = self.client.put(reverse('api:classification-validate', kwargs={'pk': classification.id}))
        self.assertEqual(response.status_code, status.HTTP_202_ACCEPTED)
        self.assertDictEqual(response.json(), {
            'id': str(classification.id),
            'source': {
                'id': str(classification.source.id),
                'type': classification.source.type.value,
                'slug': classification.source.slug,
                'revision': classification.source.revision,
                'internal': classification.source.internal
            },
            'ml_class': {
                'id': str(classification.ml_class.id),
                'name': classification.ml_class.name
            },
            'state': ClassificationState.Validated.value,
            'confidence': classification.confidence
        })
