from enum import Enum
from rest_framework import serializers
from rest_framework.schemas.openapi import AutoSchema as BaseAutoSchema
import warnings


class AutoSchema(BaseAutoSchema):
    """
    A custom view schema generator.
    The docs on OpenAPI generation are currently very incomplete.
    If you need to implement more features to avoid the "patch.py" and allow
    views to customize their OpenAPI schemas by themselves, you may see for
    yourself how DRF does the schema generation:
    https://github.com/encode/django-rest-framework/blob/master/rest_framework/schemas/openapi.py
    """

    def _map_serializer(self, serializer):
        """
        This is a temporary patch because the default schema generation does not handle
        callable defaults in fields properly, and adds 'default=null' to any field that
        does not have a default value, even required fields.

        https://github.com/encode/django-rest-framework/issues/6858
        """
        schema = super()._map_serializer(serializer)
        for field_name in schema['properties']:
            if 'default' not in schema['properties'][field_name]:
                continue
            default = schema['properties'][field_name]['default']

            if hasattr(default, 'openapi_value'):
                # Allow a 'openapi_value' attribute on the default for custom defaults
                default = default.openapi_value

            elif callable(default):
                # Try to call the callable default; if it does not work, warn, then remove it
                try:
                    default = default()
                except Exception as e:
                    warnings.warn('Unsupported callable default for field {}: {!s}'.format(field_name, e))
                    del schema['properties'][field_name]['default']
                    continue

            elif isinstance(default, Enum):
                # Convert enums into their string values
                default = default.value

            # Remove null defaults on required fields
            if default is None and field_name in schema.get('required', []):
                del schema['properties'][field_name]['default']
            else:
                schema['properties'][field_name]['default'] = default

        return schema

    def _map_field(self, field):
        """
        Yet another temporary patch because HStoreField is not properly treated as `type: object`.

        https://github.com/encode/django-rest-framework/issues/6913
        https://github.com/encode/django-rest-framework/pull/6914
        """
        schema = super()._map_field(field)
        if isinstance(field, serializers.HStoreField):
            assert schema['type'] == 'string'  # Will crash when this is fixed upstream
            return {
                'type': 'object'
            }
        return schema

    def get_operation(self, path, method):
        operation = super().get_operation(path, method)

        # Operation IDs for list endpoints are improperly cased: listThings instead of ListThings
        # https://github.com/encode/django-rest-framework/pull/6917
        if operation['operationId'][0].islower():
            operation['operationId'] = operation['operationId'][0].upper() + operation['operationId'][1:]

        # Setting deprecated = True on a View makes it deprecated in OpenAPI
        if getattr(self.view, 'deprecated', False):
            operation['deprecated'] = True

        # Allow an `openapi_overrides` attribute to override the operation's properties
        # TODO: Quite crude, not enough to make overriding request/response bodies easy
        if hasattr(self.view, 'openapi_overrides'):
            operation.update(self.view.openapi_overrides)

        return operation
